from __future__ import annotations

import logging
import os
import time
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable

import numpy as np
import psutil

if TYPE_CHECKING:
    import numpy.typing as npt

logger = logging.getLogger(__name__)


def steplength_exponential(
    iteration: int,
    min_steplength: float = 0.3,
    max_steplength: float = 0.6,
    halflife: float = 1.5,
) -> float:
    r"""
    Compute a suitable step length for the update step.

    This is an implementation of Eq. (49), which calculates a suitable step length for
    the update step, from the book: \"Formulating the history matching problem with
    consistent error statistics", written by :cite:t:`evensen2021formulating`.

    Examples
    --------
    >>> [steplength_exponential(i) for i in [1, 2, 3, 4]]
    [0.6, 0.48898815748423097, 0.41905507889761495, 0.375]
    >>> [steplength_exponential(i, 0.0, 1.0, 1.0) for i in [1, 2, 3, 4]]
    [1.0, 0.5, 0.25, 0.125]
    >>> [steplength_exponential(i, 0.0, 1.0, 0.5) for i in [1, 2, 3, 4]]
    [1.0, 0.25, 0.0625, 0.015625]
    >>> [steplength_exponential(i, 0.5, 1.0, 1.0) for i in [1, 2, 3]]
    [1.0, 0.75, 0.625]

    """
    assert max_steplength > min_steplength
    assert iteration >= 1
    assert halflife > 0

    delta = max_steplength - min_steplength
    exponent = -(iteration - 1) / halflife
    return min_steplength + delta * 2**exponent


def _validate_inputs(
    parameters: npt.NDArray[np.double],
    covariance: npt.NDArray[np.double],
    observations: npt.NDArray[np.double],
) -> None:
    # Check types
    inputs = [parameters, covariance, observations]
    names = ["parameters", "covariances", "observations"]
    for input_, name in zip(inputs, names):
        if not isinstance(input_, np.ndarray):
            raise TypeError(f"Argument '{name}' must be numpy nd.array")

    assert covariance.ndim in (1, 2)
    assert parameters.ndim == 2

    assert covariance.shape[0] == observations.shape[0]
    assert covariance.shape[0] == observations.shape[0]


def sample_mvnormal(
    *,
    C_dd_cholesky: npt.NDArray[np.double],
    rng: np.random._generator.Generator,
    size: int,
) -> npt.NDArray[np.double]:
    """Draw samples from the multivariate normal N(0, C_dd).

    We write this function from scratch to avoid factoring the covariance
    matrix every time we sample, and we want to exploit diagonal covariance
    matrices in terms of computation and memory. More specifically:

        - numpy.random.multivariate_normal factors the covariance in every call
        - scipy.stats.Covariance.from_diagonal stores off diagonal zeros

    So the best choice was to write sampling from scratch.


    Examples
    --------
    >>> C_dd_cholesky = np.diag([5, 10, 15])
    >>> rng = np.random.default_rng(42)
    >>> sample_mvnormal(C_dd_cholesky=C_dd_cholesky, rng=rng, size=2)
    array([[  1.5235854 ,  -5.19992053],
           [  7.50451196,   9.40564716],
           [-29.26552783, -19.5326926 ]])
    >>> sample_mvnormal(C_dd_cholesky=np.diag(C_dd_cholesky), rng=rng, size=2)
    array([[ 0.63920202, -1.58121296],
           [-0.16801158, -8.53043928],
           [13.19096962, 11.66687903]])


    """
    # Standard normal samples
    z = rng.standard_normal(size=(C_dd_cholesky.shape[0], size))

    # A 2D covariance matrix was passed
    if C_dd_cholesky.ndim == 2:
        return C_dd_cholesky @ z

    # A 1D diagonal of a covariance matrix was passed
    return C_dd_cholesky.reshape(-1, 1) * z


def memory_usage_decorator(
    enabled: bool = True,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def wrapper(*args, **kwargs) -> Any:  # type: ignore
            if enabled:
                # Get process ID of the current Python process
                process = psutil.Process(os.getpid())

                start_time = time.perf_counter()

                # Memory usage before the function call
                mem_before = process.memory_info().rss / 1024 / 1024  # Convert to MB
                logger.debug(
                    f"\nMemory before calling '{func.__name__}': {mem_before:.2f} MB"
                )

            # Call the target function
            result = func(*args, **kwargs)

            if enabled:
                mem_after = process.memory_info().rss / 1024 / 1024  # Convert to MB
                end_time = time.perf_counter()
                # Memory usage after the function call
                logger.debug(
                    f"Memory after calling '{func.__name__}': {mem_after:.2f} MB"
                )
                logger.debug(f"Run time used: {(end_time - start_time):.4f} seconds")
            return result

        return wrapper

    return decorator


def calc_max_number_of_layers_per_batch_for_distance_localization(
    nx: int,
    ny: int,
    nz: int,
    num_obs: int,
    nreal: int,
    bytes_per_float: int = 8,  # float64 as default here
) -> int:
    """Calculate number of layers from a 3D field parameter that can be updated
    within available memory. Distance-based localization requires two large matrices
    the Kalman gain matrix K and the localization scaling matrix RHO, both have size
    equal to number of field parameter values times number of observations.
    Therefore, a batching algorithm is used where only a subset of parameters
    is used when calculating the Schur product of RHO and K matrix in the update
    algorithm. This function calculates number of batches and
    number of grid layers of field parameter values that can fit
    into the available memory for one batch accounting for a safety margin.

    The available memory is checked using the `psutil` library, which provides
    information about system memory usage.
    From `psutil` documentation:
    - available:
        the memory that can be given instantly to processes without the
        system going into swap.
        This is calculated by summing different memory values depending
        on the platform and it is supposed to be used to monitor actual
        memory usage in a cross platform fashion.

    Args:
        nx: grid size in I-direction (local x-axis direction)
        ny: grid size in J-direction (local y-axis direction)
        nz: grid size in K-direction (number of layers)
        num_obs: Number of observations
        nreal: Number of realizations
        bytes_per_float: Is 4 or 8

    Returns:
        Max number of layers that can be updated in one batch to
        avoid memory problems.

    """

    memory_safety_factor = 0.8
    num_params = nx * ny * nz
    num_param_per_layer = nx * ny

    # Rough estimate of necessary number of float variables
    sum_floats = 0
    sum_floats += num_params * num_obs  # K matrix before Schur product
    sum_floats += num_params * num_obs  # RHO matrix
    sum_floats += num_params * num_obs  # K matrix after Schur product
    sum_floats += int(num_params * nreal * 2.5)  # X_prior, X_prior_batch, M_delta
    sum_floats += int(num_params * nreal * 1.5)  # X_post and X_post_batch
    sum_floats += num_obs * nreal * 2  # D matrix and internal matrices
    sum_floats += num_obs * nreal * 2  # Y matrix and internal matrices

    # Check available memory
    available_memory_in_bytes = psutil.virtual_memory().available * memory_safety_factor

    # Required memory
    total_required_memory_per_field_param = sum_floats * bytes_per_float

    # Minimum number of batches
    min_number_of_batches = int(
        np.ceil(total_required_memory_per_field_param / available_memory_in_bytes)
    )

    max_nlayer_per_batch = int(nz / min_number_of_batches)

    if max_nlayer_per_batch == 0:
        # Batch size cannot be less than 1 layer
        memory_one_batch = num_param_per_layer * bytes_per_float
        raise MemoryError(
            "The required memory to update one grid layer or one 2D surface is "
            "larger than available memory.\n"
            "Cannot split the update into batch size less than one complete "
            "grid layer for 3D field or one surface for 2D fields."
            f"Required memory for one batch is about: {memory_one_batch / 10**9} GB\n"
            f"Available memory is about: {available_memory_in_bytes / 10**9} GB"
        )

    logger.debug(
        "Calculate batch size for updating of field parameter:\n"
        f" Number of parameters in field param: {num_params}\n"
        f" Required number of floats to update one field parameter: {sum_floats}\n"
        " Available memory per field param update: "
        f"{available_memory_in_bytes / 10**9} GB\n"
        " Required memory total to update a field parameter: "
        f"{total_required_memory_per_field_param / 10**9} GB\n"
        f" Number of layers in one batch: {max_nlayer_per_batch}"
    )

    return max_nlayer_per_batch


def transform_positions_to_local_field_coordinates(
    coordsys_origin: tuple[float, float],
    coordsys_rotation_angle: float,
    utmx: npt.NDArray[np.float64],
    utmy: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """Calculates coordinate transformation from global to local coordinates.

    Args:
        coordys_origin: (x,y) coordinate of local coordinate
        origin in global coordinates.
        coordsys_rotation_angle: Angle for how much the local x-axis is rotated
        anti-clockwise relative to the global x-axis in degrees.
        utmx: vector of x-coordinates in global coordinates.
        utmy: vector of y-coordinates in global coordinates.

    Returns:
        First vector is local x-coordinates and second vector is local y-coordinates.
    """
    # Translate
    x1 = utmx - coordsys_origin[0]
    y1 = utmy - coordsys_origin[1]
    # Rotate
    # Input angle is the local coordinate systems rotation
    # anticlockwise relative to global x-axis in degrees
    rotation_of_ertbox = coordsys_rotation_angle
    rotation_angle = np.deg2rad(rotation_of_ertbox)
    cos_theta = np.cos(rotation_angle)
    sin_theta = np.sin(rotation_angle)
    x2 = x1 * cos_theta + y1 * sin_theta
    y2 = -x1 * sin_theta + y1 * cos_theta
    return x2, y2


def transform_local_ellipse_angle_to_local_coords(
    coordsys_rotation_angle: float,
    ellipse_anisotropy_angle: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Calculate angles relative to local coordinate system.

    Args:
        coordsys_rotation_angle: Local coordinate systems rotation angle
        relative to the global coordinate system.
        ellipse_anisotropy_angle: Vector of input angles in global coordinates.

    Returns:
        Vector of output angles relative to the local coordinate system.
    """
    # Both angles measured anti-clock from global coordinate systems x-axis in degrees
    return ellipse_anisotropy_angle - coordsys_rotation_angle


def localization_scaling_function(
    distances: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Calculate scaling factor to be used as values in
    RHO matrix in distance-based localization.
    The scaling function implements the commonly
    used function published by Gaspari and Cohn.
    For input normalized distance >= 2, the value will be 0.

    Args:
        distances: Vector of values for normalized distances.

    Returns:
        Values of scaling factors for each value of input distance.
    """
    # "gaspari-cohn"
    # Commonly used in distance-based localization
    # Is exact 0 for normalized distance > 2.
    scaling_factor = distances
    d2 = distances**2
    d3 = d2 * distances
    d4 = d3 * distances
    d5 = d4 * distances
    s = -1 / 4 * d5 + 1 / 2 * d4 + 5 / 8 * d3 - 5 / 3 * d2 + 1
    scaling_factor[distances <= 1] = s[distances <= 1]
    s = (
        1 / 12 * d5
        - 1 / 2 * d4
        + 5 / 8 * d3
        + 5 / 3 * d2
        - 5 * distances
        + 4
        - 2 / 3 * 1 / distances
    )
    scaling_factor[(distances > 1) & (distances <= 2)] = s[
        (distances > 1) & (distances <= 2)
    ]
    scaling_factor[distances > 2] = 0.0

    return scaling_factor


def calc_rho_for_2d_grid_layer(
    nx: int,
    ny: int,
    xinc: float,
    yinc: float,
    obs_xpos: npt.NDArray[np.float64],
    obs_ypos: npt.NDArray[np.float64],
    obs_main_range: npt.NDArray[np.float64],
    obs_perp_range: npt.NDArray[np.float64],
    obs_anisotropy_angle: npt.NDArray[np.float64],
    right_handed_grid_indexing: bool = True,
) -> npt.NDArray[np.float64]:
    """Calculate scaling values (RHO matrix elements) for a set of observations
    with associated localization ellipse. The method will first
    calculate the distances from each observation position to each grid cell
    center point of all grid cells for a 2D grid.
    The localization method will only consider lateral distances, and it is
    therefore sufficient to calculate the distances in 2D.
    All input observation positions are in the local grid coordinate system
    to simplify the calculation of the distances.

    The position: xpos[n], ypos[n] and
    localization ellipse defined by obs_main_range[n],obs_perp_range[n],
    obs_anisotropy_angle[n]) refers to observation[n].

    The distance between an observation with index n and a grid cell (i,j) is
    d[m,n] = dist((xpos_obs[n],ypos_obs[n]),(xpos_field[i,j],ypos_field[i,j]))

    RHO[[m,n] = scaling(d)
    where m = j + i * ny for left-handed grid index origo and
          m = (ny - j - 1) + i * ny for right-handed grid index origo
    Note that since d[m,n] does only depend on observation index n and
    grid cell index (i,j). The values for RHO is
    calculated for the combination ((i,j), n) and this covers
    one grid layer in ertbox grid or a 2D surface grid.

    Args:
        nx: Number of grid cells in x-direction of local coordinate system.
        ny: Number of grid cells in y-direction of local coordinate system.
        xinc: Grid cell size in x-direction.
        yinc: Grid cell size in y-direction.
        obs_xpos: Observations x coordinates in local coordinates
        obs_ypos: Observatiopns y coordinates in local coordinates
        obs_main_range: Localization ellipse first range
        obs_perp_range: Localization ellipse second range
        obs_anisotropy_angle: Localization ellipse orientation relative
        to local coordinate system in degrees

    Returns:
        Rho matrix values for one layer of the 3D ertbox grid or for a 2D surface grid.
    """
    # Center points of each grid cell in field parameter grid
    x_local = (np.arange(nx, dtype=np.float64) + 0.5) * xinc
    if right_handed_grid_indexing:
        # y coordinate decreases from max to min
        y_local = (np.arange(ny - 1, -1, -1, dtype=np.float64) + 0.5) * yinc
    else:
        # y coordinate increases from min to max
        y_local = (np.arange(ny, dtype=np.float64) + 0.5) * yinc
    mesh_x_coord, mesh_y_coord = np.meshgrid(x_local, y_local, indexing="ij")

    # Number of observations
    nobs = len(obs_xpos)
    assert nobs == len(obs_ypos), (
        "Number of coordinates must match number of observations"
    )
    assert nobs == len(obs_anisotropy_angle), (
        "Number of ellipse orientation angles must match number of observations"
    )
    assert nobs == len(obs_main_range), (
        "Number of ellipse main range values must match number of observations"
    )
    assert nobs == len(obs_perp_range), (
        "Number of ellipse second range values must match number of observations"
    )
    assert np.all(obs_main_range > 0.0), (
        "All range values for all observations must be positive"
    )
    assert np.all(obs_perp_range > 0.0), (
        "All range values for all observations must be positive"
    )

    # Expand grid coordinates to match observations
    mesh_x_coord_flat = mesh_x_coord.flatten()[:, np.newaxis]  # (nx * ny, 1)
    mesh_y_coord_flat = mesh_y_coord.flatten()[:, np.newaxis]  # (nx * ny, 1)

    # Observation coordinates and parameters
    obs_xpos = obs_xpos[np.newaxis, :]  # (1, nobs)
    obs_ypos = obs_ypos[np.newaxis, :]  # (1, nobs)
    obs_main_range = obs_main_range[np.newaxis, :]  # (1, nobs)
    obs_perp_range = obs_perp_range[np.newaxis, :]  # (1, nobs)
    obs_anisotropy_angle = obs_anisotropy_angle[np.newaxis, :]  # (1, nobs)

    # Compute displacement between grid points and observations
    dX = mesh_x_coord_flat - obs_xpos  # (nx * ny, nobs)
    dY = mesh_y_coord_flat - obs_ypos  # (nx * ny, nobs)

    # Compute rotation parameters
    rotation = np.deg2rad(obs_anisotropy_angle)
    cos_angle = np.cos(rotation)  # (1, nobs)
    sin_angle = np.sin(rotation)  # (1, nobs)

    # Rotate and scale displacements to local coordinate system defined
    # by the two half axes of the influence ellipse. First coordinate (local x) is in
    # direction defined by anisotropy angle and local y is perpendicular to that.
    # Scale the distance by the ranges to get a normalized distance
    # (with value 1 at the edge of the ellipse)
    dX_ellipse = (dX * cos_angle + dY * sin_angle) / obs_main_range  # (nx * ny, nobs)
    dY_ellipse = (-dX * sin_angle + dY * cos_angle) / obs_perp_range  # (nx * ny, nobs)

    # Compute distances in the elliptical coordinate system
    distances = np.hypot(dX_ellipse, dY_ellipse)  # (nx * ny, nobs)
    # Apply the scaling function
    return localization_scaling_function(distances).reshape((nx, ny, nobs))


if __name__ == "__main__":
    import pytest

    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(args=[__file__, "--doctest-modules", "-v", "-v"])



import functools
import sys
import time
from copy import deepcopy

import numpy as np
import pytest
from numpy import typing as npt
from numpy.testing import assert_allclose

from iterative_ensemble_smoother import ESMDA
from iterative_ensemble_smoother.esmda_inversion import (
    empirical_cross_covariance,
    normalize_alpha,
)
from iterative_ensemble_smoother.experimental import (
    AdaptiveESMDA,
    DistanceESMDA,
    ensemble_smoother_update_step_row_scaling,
)
from iterative_ensemble_smoother.utils import (
    calc_max_number_of_layers_per_batch_for_distance_localization,
    calc_rho_for_2d_grid_layer,
)


@pytest.fixture
def linear_problem(request):
    # Seed the problem using indirect parametrization:
    # https://docs.pytest.org/en/latest/example/parametrize.html#indirect-parametrization
    rng = np.random.default_rng(request.param)

    # Create a problem with g(x) = A @ x
    num_parameters = 50
    num_observations = 10
    num_ensemble = 200

    A = np.exp(rng.standard_normal(size=(num_observations, num_parameters)))

    def g(X):
        """Forward model."""
        return A @ X

    # Create observations
    x_true = np.linspace(-1, 1, num=num_parameters)
    observations = g(x_true) + rng.standard_normal(size=num_observations) / 10

    # Initial ensemble and covariance
    X = rng.normal(size=(num_parameters, num_ensemble))
    covariance = rng.triangular(0.1, 1, 1, size=num_observations)
    return X, g, observations, covariance, rng


class TestAdaptiveESMDA:
    @pytest.mark.parametrize(
        "linear_problem",
        range(25),
        indirect=True,
        ids=[f"seed-{i + 1}" for i in range(25)],
    )
    def test_that_adaptive_localization_with_cutoff_1_equals_ensemble_prior(
        self, linear_problem
    ):
        # Create a problem with g(x) = A @ x
        X, g, observations, covariance, _ = linear_problem

        # Create adaptive smoother
        smoother = AdaptiveESMDA(
            covariance=covariance, observations=observations, seed=1
        )

        def correlation_callback(corr_matrix):
            # cross-correlation matrix contains all correlations,
            # even those deemed insignificant.
            print(corr_matrix)
            assert corr_matrix.shape[0] == 50
            assert corr_matrix.shape[1] == len(observations)

        X_i = np.copy(X)
        alpha = normalize_alpha(np.ones(5))
        for alpha_i in alpha:
            # Run forward model
            Y_i = g(X_i)

            # Create noise D - common to this ESMDA update
            D_i = smoother.perturb_observations(
                ensemble_size=Y_i.shape[1], alpha=alpha_i
            )

            # Update the relevant parameters and write to X
            X_i = smoother.assimilate(
                X=X_i,
                Y=Y_i,
                D=D_i,
                alpha=alpha_i,
                correlation_threshold=1,
                correlation_callback=correlation_callback,
            )

        assert np.allclose(X, X_i)

    @pytest.mark.parametrize(
        "linear_problem",
        range(25),
        indirect=True,
        ids=[f"seed-{i + 1}" for i in range(25)],
    )
    def test_that_adaptive_localization_with_cutoff_0_equals_standard_ESMDA_update(
        self, linear_problem
    ):
        # Create a problem with g(x) = A @ x
        X, g, observations, covariance, _ = linear_problem

        # =============================================================================
        # SETUP ESMDA FOR LOCALIZATION AND SOLVE PROBLEM
        # =============================================================================
        alpha = normalize_alpha(np.ones(5))

        smoother = AdaptiveESMDA(
            covariance=covariance, observations=observations, seed=1
        )

        def correlation_callback(corr_matrix):
            # A correlation threshold of 0 means that all
            # correlations are deemed significant.
            # Therefore, the cross-correlation matrix must
            # include all parameter-resposne pairs.
            assert corr_matrix.shape[0] == X.shape[0]
            assert corr_matrix.shape[1] == len(observations)

        X_i = np.copy(X)
        for _, alpha_i in enumerate(alpha, 1):
            # Run forward model
            Y_i = g(X_i)

            # Create noise D - common to this ESMDA update
            D_i = smoother.perturb_observations(
                ensemble_size=Y_i.shape[1], alpha=alpha_i
            )

            # Update the relevant parameters and write to X
            smoother.assimilate(
                X=X_i,
                Y=Y_i,
                D=D_i,
                overwrite=True,
                alpha=alpha_i,
                correlation_threshold=0,
                correlation_callback=functools.partial(correlation_callback),
            )

        # =============================================================================
        # VERIFY RESULT AGAINST NORMAL ESMDA ITERATIONS
        # =============================================================================
        smoother = ESMDA(
            covariance=covariance,
            observations=observations,
            alpha=alpha,
            seed=1,
        )

        X_i2 = np.copy(X)
        for _ in range(smoother.num_assimilations()):
            X_i2 = smoother.assimilate(X_i2, g(X_i2))

        assert np.allclose(X_i, X_i2)

    @pytest.mark.parametrize(
        "linear_problem",
        range(25),
        indirect=True,
        ids=[f"seed-{i + 1}" for i in range(25)],
    )
    @pytest.mark.parametrize(
        "cutoffs", [(0, 1e-3), (0.1, 0.2), (0.5, 0.5 + 1e-12), (0.9, 1), (1 - 1e-3, 1)]
    )
    @pytest.mark.parametrize("full_covariance", [True, False])
    def test_that_posterior_generalized_variance_increases_in_cutoff(
        self, linear_problem, cutoffs, full_covariance
    ):
        """This property only holds in the limit as the number of
        ensemble members goes to infinity. As the number of ensemble
        members decrease, this test starts to fail more often."""

        # Create a problem with g(x) = A @ x
        X, g, observations, covariance, _ = linear_problem
        if full_covariance:
            covariance = np.diag(covariance)

        # =============================================================================
        # SETUP ESMDA FOR LOCALIZATION AND SOLVE PROBLEM
        # =============================================================================
        alpha = normalize_alpha(np.ones(1))

        smoother = AdaptiveESMDA(
            covariance=covariance, observations=observations, seed=1
        )

        X_i = np.copy(X)
        for _, alpha_i in enumerate(alpha, 1):
            # Run forward model
            Y_i = g(X_i)

            # Create noise D - common to this ESMDA update
            D_i = smoother.perturb_observations(
                ensemble_size=Y_i.shape[1], alpha=alpha_i
            )

            cutoff_low, cutoff_high = cutoffs
            assert cutoff_low <= cutoff_high

            X_i_low_cutoff = smoother.assimilate(
                X=X_i,
                Y=Y_i,
                D=D_i,
                alpha=alpha_i,
                correlation_threshold=cutoff_low,
            )
            X_i_high_cutoff = smoother.assimilate(
                X=X_i,
                Y=Y_i,
                D=D_i,
                alpha=alpha_i,
                correlation_threshold=cutoff_high,
            )

            # Compute covariances
            prior_cov = np.cov(X, rowvar=True)
            posterior_cutoff_low_cov = np.cov(X_i_low_cutoff, rowvar=True)
            posterior_cutoff_high_cov = np.cov(X_i_high_cutoff, rowvar=True)
            assert prior_cov.shape == (X.shape[0], X.shape[0])

            # Compute determinants of covariance matrices
            # https://en.wikipedia.org/wiki/Generalized_variance
            # intuitively: large determintant => high covariancce
            #  => larger volume of multivariate normal
            # => less information contained in multivariate normal
            generalized_variance_prior = np.linalg.det(prior_cov)
            generalized_variance_low = np.linalg.det(posterior_cutoff_low_cov)
            generalized_variance_high = np.linalg.det(posterior_cutoff_high_cov)

            # The covariance is positive (semi) definite, so the determinant is >= 0
            assert generalized_variance_low >= 0, f"1 Failed with cutoff={cutoffs}"

            # Assimilating information always leads to more information,
            # which means a smaller lower determinant
            assert generalized_variance_low <= generalized_variance_prior
            assert generalized_variance_high <= generalized_variance_prior

            # A higher threshold means we assimilate less information
            assert generalized_variance_low <= generalized_variance_high, (
                f"2 Failed with cutoff_low={cutoff_low} and cutoff_high={cutoff_high}"
            )

    @pytest.mark.parametrize(
        "linear_problem",
        range(25),
        indirect=True,
        ids=[f"seed-{i + 1}" for i in range(25)],
    )
    def test_that_adaptive_localization_works_with_dying_realizations(
        self,
        linear_problem,
    ):
        """A full worked example of the adaptive localization API, with
        parameter groups and dying realizations (if compute nodes go down).

        This is mainly meant as an example, not a test. To make it into a test
        worth running, we set the threshold for the correlation to 0 and verify
        that it returns the same result as ESMDA."""

        def zero_correlation_threshold(ensemble_size):
            """Adaptive localization only matches ESMDA when the threshold is <=0."""
            return 0

        # Create a problem with g(x) = A @ x
        X, g, observations, covariance, rng = linear_problem
        num_parameters, num_ensemble = X.shape
        num_observations = len(observations)

        # Split parameters into groups of equal size
        num_groups = 10
        assert num_observations % num_groups == 0, "Num groups must divide parameters"
        parameters_groups = np.array_split(np.arange(num_parameters), num_groups)
        assert len(parameters_groups) == num_groups

        # =============================================================================
        # SETUP ESMDA FOR LOCALIZATION AND SOLVE PROBLEM
        # =============================================================================
        alpha = normalize_alpha(np.array([5, 4, 3, 2, 1]))  # Vector of inflation values
        start_time = time.perf_counter()
        smoother = AdaptiveESMDA(
            covariance=covariance, observations=observations, seed=1
        )

        # Simulate realizations that die
        living_mask = rng.choice(
            [True, False], size=(len(alpha), num_ensemble), p=[0.9, 0.1]
        )

        X_i = np.copy(X)
        for i, alpha_i in enumerate(alpha, 1):
            print(f"ESMDA iteration {i} with alpha_i={alpha_i}")

            # Run forward model
            Y_i = g(X_i)

            # We simulate loss of realizations due to compute clusters going down.
            # Figure out which realizations are still alive:
            alive_mask_i = np.all(living_mask[:i, :], axis=0)
            num_alive = alive_mask_i.sum()
            print(f"  Total realizations still alive: {num_alive} / {num_ensemble}")

            # Create noise D - common to this ESMDA update
            D_i = smoother.perturb_observations(ensemble_size=num_alive, alpha=alpha_i)

            # Loop over parameter groups and update
            for j, parameter_mask_j in enumerate(parameters_groups, 1):
                print(f"  Updating parameter group {j}/{len(parameters_groups)}")

                # Mask out rows in this parameter group, and columns of realization
                # that are still alive. This step simulates fetching from disk.
                mask = np.ix_(parameter_mask_j, alive_mask_i)

                # Update the relevant parameters and write to X
                X_i[mask] = smoother.assimilate(
                    X=X_i[mask],
                    Y=Y_i[:, alive_mask_i],
                    D=D_i,
                    alpha=alpha_i,
                    # Pass a function that always returns zero,
                    # no matter what the ensemble size is
                    correlation_threshold=zero_correlation_threshold,
                )

            print()

        print(f"ESMDA with localization - Ran in {time.perf_counter() - start_time} s")

        # =============================================================================
        # VERIFY RESULT AGAINST NORMAL ESMDA ITERATIONS
        # =============================================================================
        start_time = time.perf_counter()
        smoother = ESMDA(
            covariance=covariance,
            observations=observations,
            alpha=alpha,
            seed=1,
        )

        X_i2 = np.copy(X)
        for i in range(smoother.num_assimilations()):
            # Run simulations
            Y_i = g(X_i2)

            # We simulate loss of realizations due to compute clusters going down.
            # Figure out which realizations are still alive:
            alive_mask_i = np.all(living_mask[: i + 1, :], axis=0)
            num_alive = alive_mask_i.sum()

            X_i2[:, alive_mask_i] = smoother.assimilate(
                X_i2[:, alive_mask_i], Y_i[:, alive_mask_i]
            )

        # For this test to pass, correlation_threshold() should return <= 0
        print(
            "Norm difference between ESMDA with and without localization:",
            np.linalg.norm(X_i - X_i2),
        )
        assert np.allclose(X_i, X_i2, atol=1e-4)

        print(
            f"ESMDA without localization - Ran in {time.perf_counter() - start_time} s"
        )

        print("------------------------------------------")

    @pytest.mark.parametrize(
        "linear_problem",
        range(25),
        indirect=True,
        ids=[f"seed-{i + 1}" for i in range(25)],
    )
    def test_that_cov_YY_can_be_computed_outside_of_assimilate(
        self,
        linear_problem,
    ):
        """Cov(Y, Y) may be computed once in each assimilation round.
        This saves time if the user wants to iterate over parameters groups,
        since Cov(Y, Y) is independent of the parameters X, there is no reason
        to compute it more than once.

        Below we do not loop over parameter groups in X, so there is no speed
        gain when passing the covariance cov(Y, Y). This test is just to check
        that the result is the same regardless of whether the user passes
        the covariance matrix or not.
        """

        # Create a problem with g(x) = A @ x
        X, g, observations, covariance, _ = linear_problem
        _, ensemble_size = X.shape

        alpha = normalize_alpha(np.array([5, 4, 3, 2, 1]))  # Vector of inflation values
        smoother = AdaptiveESMDA(
            covariance=covariance, observations=observations, seed=1
        )

        X_i = np.copy(X)
        for i, alpha_i in enumerate(alpha, 1):
            print(f"ESMDA iteration {i} with alpha_i={alpha_i}")

            # Run forward model
            Y_i = g(X_i)

            # Create noise D - common to this ESMDA update
            D_i = smoother.perturb_observations(
                ensemble_size=ensemble_size, alpha=alpha_i
            )

            # Update the parameters without using pre-computed cov_YY
            X_i1 = smoother.assimilate(
                X=X_i,
                Y=Y_i,
                D=D_i,
                alpha=alpha_i,
            )

            # Update the parameters using pre-computed cov_YY
            X_i2 = smoother.assimilate(
                X=X_i,
                Y=Y_i,
                D=D_i,
                alpha=alpha_i,
                cov_YY=empirical_cross_covariance(Y_i, Y_i),
            )

            # Check that the update is the same, whether or not Cov_YY is passed
            assert np.allclose(X_i1, X_i2)

            X_i = X_i1

    @pytest.fixture
    def large_linear_problem(self):
        """
        Creates a problem large enough to realistically benchmark parallelization.
        """
        rng = np.random.default_rng(42)  # Use a fixed seed for reproducibility

        num_parameters = 5000
        num_observations = 200
        num_ensemble = 200

        A = np.exp(rng.standard_normal(size=(num_observations, num_parameters)))

        def g(X):
            """Forward model."""
            return A @ X

        x_true = np.linspace(-1, 1, num=num_parameters)
        observations = g(x_true) + rng.standard_normal(size=num_observations) / 10

        X = rng.normal(size=(num_parameters, num_ensemble))
        covariance = rng.triangular(0.1, 1, 1, size=num_observations)
        return X, g, observations, covariance

    def test_parallelization_runtime_comparison(self, large_linear_problem):
        """
        Compares the runtime of the assimilate method with and without parallelization.
        """
        X, g, observations, covariance = large_linear_problem

        # Use a single assimilation step for a clear comparison
        alpha = 1.0

        # --- Setup the smoother and common data ---
        smoother = AdaptiveESMDA(
            covariance=covariance, observations=observations, seed=1
        )
        Y = g(X)
        D = smoother.perturb_observations(ensemble_size=Y.shape[1], alpha=alpha)

        # Make separate copies for each run to ensure a fair start
        X_serial = np.copy(X)
        X_parallel = np.copy(X)

        # --- 1. Serial Execution ---
        print("\n--- Running Serial Benchmark (n_jobs=1) ---")
        start_serial = time.perf_counter()
        X_serial = smoother.assimilate(
            X=X_serial,
            Y=Y,
            D=D,
            alpha=alpha,
            correlation_threshold=0,  # Force the code to update every single parameter.
            n_jobs=1,
        )
        end_serial = time.perf_counter()
        time_serial = end_serial - start_serial
        print(f"Serial execution time: {time_serial:.4f} seconds")

        # --- 2. Parallel Execution ---
        print("\n--- Running Parallel Benchmark (n_jobs=-1) ---")
        start_parallel = time.perf_counter()
        X_parallel = smoother.assimilate(
            X=X_parallel,
            Y=Y,
            D=D,
            alpha=alpha,
            correlation_threshold=0,  # Force the code to update every single parameter.
            n_jobs=-1,
        )
        end_parallel = time.perf_counter()
        time_parallel = end_parallel - start_parallel
        print(f"Parallel execution time: {time_parallel:.4f} seconds")

        # --- 3. Verification and Summary ---
        print("\n--- Benchmark Summary ---")
        # Critical check: ensure both methods produce the same result
        assert np.allclose(X_serial, X_parallel)
        print("✅ Numerical results are identical.")

        if time_parallel > 0:
            speedup = time_serial / time_parallel
            print(f"🚀 Speedup factor: {speedup:.2f}x")

        assert time_serial > time_parallel


class TestRowScaling:
    def test_that_row_scaling_updates_parameters(self):
        class RowScaling:
            # Illustration of how row scaling works, `multiply` is the important part
            # For the actual implementation, which is more involved, see:
            # https://github.com/equinor/ert/blob/84aad3c56e0e52be27b9966322d93dbb85024c1c/src/clib/lib/enkf/row_scaling.cpp
            def __init__(self, alpha=1.0):
                """Alpha is the strength of the update."""
                assert 0 <= alpha <= 1.0
                self.alpha = alpha

            def multiply(self, X, K):
                """Takes a matrix X and a matrix K and performs alpha * X @ K."""
                # This implementation merely mimics how RowScaling::multiply behaves
                # in the C++ code. It mutates the input argument X instead of returning.
                X[:, :] = X @ (K * self.alpha)

        # Example showing how to use row scaling
        num_parameters = 100
        num_observations = 20
        num_ensemble = 10

        rng = np.random.default_rng(42)

        X = rng.normal(size=(num_parameters, num_ensemble))
        Y = rng.normal(size=(num_observations, num_ensemble))
        covariance = np.exp(rng.normal(size=num_observations))
        observations = rng.normal(size=num_observations, loc=1)

        # Split up X into groups of parameters as needed
        row_groups = [(0,), (1, 2), (4, 5, 6), tuple(range(7, 100))]
        X_with_row_scaling = [
            (X[idx, :], RowScaling(alpha=1 / (i + 1)))
            for i, idx in enumerate(row_groups)
        ]
        # Make a copy so we can check that update happened, since input is mutated
        X_before = deepcopy(X_with_row_scaling)

        X_with_row_scaling_updated = ensemble_smoother_update_step_row_scaling(
            covariance=covariance,
            observations=observations,
            X_with_row_scaling=X_with_row_scaling,
            Y=Y,
            seed=rng,
        )

        # Check that an update happened
        assert not np.allclose(X_before[-1][0], X_with_row_scaling_updated[-1][0])

    @pytest.mark.parametrize("inversion", list(ESMDA._inversion_methods.keys()))
    @pytest.mark.parametrize("num_ensemble", [5, 25, 200])
    def test_that_row_scaling_equal_single_ES_update(self, inversion, num_ensemble):
        class RowScaling:
            # Illustration of how row scaling works, `multiply` is the important part
            # For the actual implementation, which is more involved, see:
            # https://github.com/equinor/ert/blob/84aad3c56e0e52be27b9966322d93dbb85024c1c/src/clib/lib/enkf/row_scaling.cpp
            def __init__(self, alpha=1.0):
                """Alpha is the strength of the update."""
                assert 0 <= alpha <= 1.0
                self.alpha = alpha

            def multiply(self, X, K):
                """Takes a matrix X and a matrix K and performs alpha * X @ K."""
                # This implementation merely mimics how RowScaling::multiply behaves
                # in the C++ code. It mutates the input argument X instead of returning.
                X[:, :] = X @ (K * self.alpha)

        # Create data
        num_parameters = 100
        num_observations = 20

        rng = np.random.default_rng(num_ensemble)

        X = rng.normal(size=(num_parameters, num_ensemble))
        Y = rng.normal(size=(num_observations, num_ensemble))
        covariance = np.exp(rng.normal(size=num_observations))
        observations = rng.normal(size=num_observations, loc=1)

        # Split up X into groups of parameters
        row_groups = [tuple(range(17)), tuple(range(17, 100))]
        # When alpha=1 in row scaling, we perform a full update
        X_with_row_scaling = [
            (X[idx, :], RowScaling(alpha=1)) for _, idx in enumerate(row_groups)
        ]

        # Perform an update using row scaling
        X_with_row_scaling_updated = ensemble_smoother_update_step_row_scaling(
            covariance=covariance,
            observations=observations,
            X_with_row_scaling=X_with_row_scaling,
            Y=Y,
            inversion=inversion,
            seed=1,
        )

        # Perform an update using ESMDA API
        smoother = ESMDA(
            covariance=covariance,
            observations=observations,
            alpha=1,
            inversion=inversion,
            seed=1,
        )
        for _ in range(smoother.num_assimilations()):
            X_posterior = smoother.assimilate(X, Y)

        # The result should be the same
        assert np.allclose(
            X_posterior, np.vstack([X_i for (X_i, _) in X_with_row_scaling_updated])
        )


def assert_tests_for_distance_based_localization(
    X_prior: npt.NDArray[np.float64],
    obs_index_flatten: int,
    rho: npt.NDArray[np.float64],
    true_parameters: npt.NDArray[np.float64],
    X_posterior: npt.NDArray[np.float64],
    X_posterior_global: npt.NDArray[np.float64],
    rho_min: float = 1e-6,
    atol: float = 1e-5,
):
    # Find parameters far from the observation where localization weight is near zero
    zero_weight_indices = np.where(rho < rho_min)[0]
    assert len(zero_weight_indices) > 0, (
        "No distant points found for testing localization."
    )
    # --- Assertions for Localized Smoother ---
    # Assert distant parameters are not updated
    assert np.allclose(
        X_posterior[zero_weight_indices, :], X_prior[zero_weight_indices, :], atol=atol
    )
    # Assert parameter at observation IS updated
    assert not np.allclose(
        X_posterior[obs_index_flatten, :], X_prior[obs_index_flatten, :]
    )

    prior_mean = np.mean(X_prior, axis=1)
    posterior_mean = np.mean(X_posterior, axis=1)
    prior_variance = np.var(X_prior, axis=1)
    posterior_variance = np.var(X_posterior, axis=1)

    # Assert update is maximal at the observation point
    update_magnitude = np.abs(posterior_mean - prior_mean)
    assert np.argmax(update_magnitude) == obs_index_flatten

    # Assert variance is reduced at observation, but preserved far away
    assert posterior_variance[obs_index_flatten] < prior_variance[obs_index_flatten]
    # The variance of the ensemble should NOT change where there was no update
    assert np.allclose(
        posterior_variance[zero_weight_indices], prior_variance[zero_weight_indices]
    )

    # The global update MUST change distant parameters due to spurious correlations.
    assert not np.allclose(
        X_posterior_global[zero_weight_indices, :],
        X_prior[zero_weight_indices, :],
    )

    # --- Comparison Assertions for Global Smoother ---
    posterior_mean_global = np.mean(X_posterior_global, axis=1)
    posterior_variance_global = np.var(X_posterior_global, axis=1)

    # Assert distant parameters ARE updated due to spurious correlations
    assert not np.allclose(
        X_posterior_global[zero_weight_indices, :],
        X_prior[zero_weight_indices, :],
    )

    # Assert AVERAGE variance is reduced everywhere (the sign of ensemble collapse)
    assert np.mean(posterior_variance_global[zero_weight_indices]) < np.mean(
        prior_variance[zero_weight_indices]
    )

    # Calculate Mean Squared Error (MSE) against the true parameters
    mse_localized = np.mean((posterior_mean - true_parameters) ** 2)
    mse_non_localized = np.mean((posterior_mean_global - true_parameters) ** 2)

    # --- Overall Quality Assertion (MSE) ---
    # Assert that the localized result is a more plausible estimate
    assert mse_localized < mse_non_localized


def calculate_rho_1d(
    N_m: int, obs_index: int, localization_radius: float, xinc: float = 1.0
) -> npt.NDArray[np.float64]:
    model_grid = np.arange(N_m) * xinc
    distances = np.abs(model_grid - obs_index * xinc)
    # Ordinary definition of range,
    # rho is approximately 0.05 at localization_radius
    return np.exp(-3.0 * (distances / localization_radius) ** 2).reshape(-1, 1)


def calculate_rho_2d(
    Nx: int, Ny: int, x_obs: int, y_obs: int, localization_radius: float
) -> npt.NDArray[np.float64]:
    xx, yy = np.meshgrid(np.arange(Nx), np.arange(Ny))
    distances_2d = np.sqrt((xx - x_obs) ** 2 + (yy - y_obs) ** 2)
    distances = distances_2d.flatten()
    # Ordinary definition of range,
    # rho is approximately 0.05 at localization_radius
    return np.exp(-3.0 * (distances / localization_radius) ** 2).reshape(-1, 1)


def calculate_rho_3d(
    Nx: int,
    Ny: int,
    Nz: int,
    x_obs: int,
    y_obs: int,
    z_obs: int,
    localization_radius: float,
) -> npt.NDArray[np.float64]:
    # Create 3D coordinate grids
    zz, yy, xx = np.meshgrid(np.arange(Nz), np.arange(Ny), np.arange(Nx), indexing="ij")
    # Calculate 3D Euclidean distance from every point to the observation
    distances_3d = np.sqrt((xx - x_obs) ** 2 + (yy - y_obs) ** 2 + (zz - z_obs) ** 2)
    distances = distances_3d.flatten()
    # Ordinary definition of range,
    # rho is approximately 0.05 at localization_radius
    return np.exp(-3.0 * (distances / localization_radius) ** 2).reshape(-1, 1)


def draw_1D_field(
    mean: float,
    stdev: float,
    xinc: float,
    corr_func_name: str,
    corr_range: float,
    nparam: int,
    nreal: int,
    rng,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    # Draw prior ensemble of 1D field with nparam. Ensemble size is nreal
    # Returns prior ensemble drawn and covariance matrix

    variance = stdev**2

    # Generate distance matrix
    x_coords = (np.arange(nparam) + 0.5) * xinc
    distances = np.abs(x_coords[:, None] - x_coords[None, :]) / corr_range

    # Compute covariance matrix based on correlation function
    if corr_func_name == "exponential":
        cov_matrix = variance * np.exp(-3.0 * distances)
    elif corr_func_name == "gaussian":
        cov_matrix = variance * np.exp(-3.0 * distances**2)
    else:
        raise ValueError("Unsupported correlation function")

    # Create mean array
    mean_values = np.full((nparam,), mean, dtype=np.float64)

    # Generate random fields with multivariate normal distribution
    fields = rng.multivariate_normal(mean_values, cov_matrix, size=nreal).T
    return fields, cov_matrix


def generate_alpha_vector(nalpha):
    """Generate alpha vector with length nalpha
    where sum of the inverse of alpha is 1.
    """
    # Define alpha[k] = [2**(m-1) + 2**(m-2) + ... + 2**(0)] / 2**(k)
    # such that alpha[k] = (2**m -1)/(2**k)
    # Then sum (1/alpha[k]) for k=0,1,2,..m-1  is 1.0
    # The alpha values are reduced with a factor 1/2 for each iteration
    # Special case: m = 3 gives alpha_vector = [7, 3.5, 1.75]
    return (2**nalpha - 1) / 2 ** np.arange(nalpha)


def predicted_mean_field_values(
    obs_vector: npt.NDArray[np.float64],
    obs_index_vector: npt.NDArray[np.int32],
    field_cov_matrix: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Calculates simple kriging estimate of expected value. This is used
    to check that DL-ESMDA and ordinary ESMDA are close to the kriging estimate
    when number of realizations becomes large.
    """
    # Extract the observed covariance matrix
    # (corresponding to the response covariance matrix)
    # Here simple kriging with 0 observation error is used, and
    # response is equal to the simulated field values in
    # position of the observations.
    obs_cov_matrix = field_cov_matrix[np.ix_(obs_index_vector, obs_index_vector)]

    # Compute the inverse covariance matrix
    inv_cov_matrix = np.linalg.inv(obs_cov_matrix)

    # Extract the covariance matrix between the observations and the full field
    field_obs_cov = field_cov_matrix[obs_index_vector, :]

    # Predicted mean values (broadcast calculation for all parameters at once)
    return field_obs_cov.T @ inv_cov_matrix @ obs_vector


@pytest.mark.parametrize("seed", list(range(9)))
def test_distance_based_localization_on_1D_case_with_single_observation(seed):
    N_m = 100  # Number of model parameters (grid points)
    N_e = 50  # Ensemble size
    j_obs = 50  # Index of the single observation

    alpha_i = 1
    obs_error_var = 0.01  # Variance of observation error

    true_parameters = np.zeros(N_m)

    true_observations = np.array([1.0])

    C_D = np.array([obs_error_var])

    rng = np.random.default_rng(seed)
    X_prior = rng.normal(loc=0.0, scale=0.5, size=(N_m, N_e))

    # Predict observations `Y` using the identity model `g(x) = x`
    # We only observe the state at `j_obs`.
    Y = X_prior[[j_obs], :]

    # Using a simple Gaussian decay for rho
    localization_radius = 20.0
    rho = calculate_rho_1d(N_m, j_obs, localization_radius)

    esmda_distance = DistanceESMDA(
        covariance=C_D, observations=true_observations, alpha=alpha_i, seed=rng
    )
    X_posterior = esmda_distance.assimilate(X=X_prior, Y=Y, rho=rho)

    esmda = ESMDA(
        covariance=C_D, observations=true_observations, alpha=alpha_i, seed=rng
    )
    X_posterior_global = esmda.assimilate(X=X_prior, Y=Y)
    assert_tests_for_distance_based_localization(
        X_prior,
        j_obs,
        rho,
        true_parameters,
        X_posterior,
        X_posterior_global,
        rho_min=1e-5,
        atol=1e-2,
    )


@pytest.mark.parametrize("seed", list(range(9)))
def test_distance_based_localization_on_2D_case_with_single_observation(seed):
    Nx, Ny = 10, 10  # Dimensions of the 2D parameter grid
    N_m = Nx * Ny  # Total number of model parameters
    N_e = 50  # Ensemble size
    x_obs, y_obs = 5, 5  # Index of the single observation in 2D

    alpha_i = 1
    obs_error_var = 0.01

    true_parameters = np.zeros(N_m)
    true_observations = np.array([1.0])
    C_D = np.array([obs_error_var])

    # --- Generate Initial Ensemble and Predictions ---
    rng = np.random.default_rng(seed)
    X_prior = rng.normal(loc=0.0, scale=0.5, size=(N_m, N_e))

    # Convert the 2D observation index to a flat 1D index for slicing
    flat_obs_index = y_obs * Nx + x_obs
    Y = X_prior[[flat_obs_index], :]

    # --- Construct 2D Localization `rho` ---
    localization_radius = 2.5
    rho = calculate_rho_2d(Nx, Ny, x_obs, y_obs, localization_radius)

    # --- Run Assimilations ---
    # Initialize smoothers with the current run's seeded RNG
    esmda_distance = DistanceESMDA(
        covariance=C_D, observations=true_observations, alpha=alpha_i, seed=rng
    )
    X_posterior = esmda_distance.assimilate(X=X_prior, Y=Y, rho=rho)

    esmda_global = ESMDA(
        covariance=C_D, observations=true_observations, alpha=alpha_i, seed=rng
    )
    X_posterior_global = esmda_global.assimilate(X=X_prior, Y=Y)

    assert_tests_for_distance_based_localization(
        X_prior,
        flat_obs_index,
        rho,
        true_parameters,
        X_posterior,
        X_posterior_global,
        rho_min=1e-6,
        atol=1e-5,
    )


@pytest.mark.parametrize("seed", list(range(9)))
def test_distance_based_localization_on_3D_case_with_single_observation(seed):
    # --- 1. System and Assimilation Setup (3D Grid) ---
    Nx, Ny, Nz = 10, 10, 10  # Dimensions of the 3D parameter grid
    N_m = Nx * Ny * Nz  # Total number of model parameters (now 1000)
    N_e = 50  # Ensemble size
    x_obs, y_obs, z_obs = 5, 5, 5  # Index of the single observation in 3D

    alpha_i = 1
    obs_error_var = 0.01

    true_parameters = np.zeros(N_m)
    true_observations = np.array([1.0])
    C_D = np.array([obs_error_var])

    # --- 2. Generate Initial Ensemble and Predictions ---
    rng = np.random.default_rng(seed)
    X_prior = rng.normal(loc=0.0, scale=0.5, size=(N_m, N_e))

    # Convert the 3D observation index to a flat 1D index for slicing
    flat_obs_index = (z_obs * Nx * Ny) + (y_obs * Nx) + x_obs
    Y = X_prior[[flat_obs_index], :]

    # --- 3. Construct 3D Localization `rho` ---
    # A smaller radius is used to ensure the localization effect is clear
    # and that some weights decay to near-zero within the grid boundaries.
    localization_radius = 2.5
    rho = calculate_rho_3d(Nx, Ny, Nz, x_obs, y_obs, z_obs, localization_radius)

    # --- 4. Run Assimilations ---
    # Initialize smoothers with the current run's seeded RNG
    esmda_distance = DistanceESMDA(
        covariance=C_D, observations=true_observations, alpha=alpha_i, seed=rng
    )
    X_posterior = esmda_distance.assimilate(X=X_prior, Y=Y, rho=rho)

    esmda_global = ESMDA(
        covariance=C_D, observations=true_observations, alpha=alpha_i, seed=rng
    )
    X_posterior_global = esmda_global.assimilate(X=X_prior, Y=Y)

    assert_tests_for_distance_based_localization(
        X_prior,
        flat_obs_index,
        rho,
        true_parameters,
        X_posterior,
        X_posterior_global,
        rho_min=1e-6,
        atol=1e-5,
    )


@pytest.mark.parametrize("seed", list(range(9)))
def test_distance_based_on_1D_case_multiple_data_assimilation(seed):
    N_m = 100  # Number of model parameters (grid points)
    N_e = 50  # Ensemble size
    j_obs = 50  # Index of the single observation
    nalpha = (
        3  # number of different alpha values. They satisfy that sum(1/alpha[i]  = 1
    )
    alpha_vector = generate_alpha_vector(nalpha)
    inverse_alpha_vector = 1.0 / alpha_vector

    # Consistency requirement for alpha
    assert abs(inverse_alpha_vector.sum() - 1.0) < 1e-7

    obs_error_var = 0.01  # Variance of observation error

    true_parameters = np.zeros(N_m)

    true_observations = np.array([1.0])

    C_D = np.array([obs_error_var])

    for iteration in range(len(alpha_vector)):
        alpha_i = np.array([alpha_vector[iteration]])
        if iteration == 0:
            # Draw prior
            rng = np.random.default_rng(seed)
            X_prior = rng.normal(loc=0.0, scale=0.5, size=(N_m, N_e))
            X_previous = X_prior.copy()
            X_previous_global = X_prior.copy()
            X_posterior = X_prior
            X_posterior_global = X_prior
        else:
            X_previous = X_posterior
            X_previous_global = X_posterior_global
        # Predict observations `Y` using the identity model `g(x) = x`
        # We only observe the state at `j_obs`.
        Y = X_previous[[j_obs], :]

        # Using a simple Gaussian decay for rho
        localization_radius = 20.0
        rho = calculate_rho_1d(N_m, j_obs, localization_radius)

        esmda_distance = DistanceESMDA(
            covariance=C_D, observations=true_observations, alpha=alpha_i, seed=rng
        )
        X_posterior = esmda_distance.assimilate(X=X_previous, Y=Y, rho=rho)

        esmda = ESMDA(
            covariance=C_D, observations=true_observations, alpha=alpha_i, seed=rng
        )
        X_posterior_global = esmda.assimilate(X=X_previous_global, Y=Y)
        assert_tests_for_distance_based_localization(
            X_prior,
            j_obs,
            rho,
            true_parameters,
            X_posterior,
            X_posterior_global,
            rho_min=1e-5,
            atol=1e-2,
        )


@pytest.mark.parametrize("seed", list(range(9)))
def test_distance_based_on_1D_case_multiple_data_assimilate_batch(seed):
    """
    This test is almost equal to the one above, but here the function
    assimilate and assimilate_batch are compared to check they give
    same result. The purpose of assimilate_batch is to avoid
    re-calculating steps that is not necessary when running a loop
    over batches of parameters
    """
    N_m = 100  # Number of model parameters (grid points)
    N_e = 50  # Ensemble size
    j_obs = 50  # Index of the single observation
    nalpha = 10  # Number of different sets of alpha values
    number_of_alpha_values = np.arange(1, nalpha, 1, dtype=np.int32)
    for m in number_of_alpha_values:
        alpha_vector = generate_alpha_vector(m)

        obs_error_var = 0.01  # Variance of observation error

        true_parameters = np.zeros(N_m)

        true_observations = np.array([1.0])

        C_D = np.array([obs_error_var])

        for iteration in range(len(alpha_vector)):
            alpha_i = np.array([alpha_vector[iteration]])
            if iteration == 0:
                # Draw prior
                rng = np.random.default_rng(seed)
                X_prior = rng.normal(loc=0.0, scale=0.5, size=(N_m, N_e))
                X_previous = X_prior.copy()
                X_previous_global = X_prior.copy()
                X_posterior = X_prior
                X_posterior_global = X_prior
            else:
                X_previous = X_posterior
                X_previous_global = X_posterior_global
            # Predict observations `Y` using the identity model `g(x) = x`
            # We only observe the state at `j_obs`.
            Y = X_previous[[j_obs], :]

            # Using a simple Gaussian decay for rho
            localization_radius = 20.0
            rho = calculate_rho_1d(N_m, j_obs, localization_radius)

            # Use same seed for both instances of DistanceESMDA
            esmda_distance = DistanceESMDA(
                covariance=C_D, observations=true_observations, alpha=alpha_i, seed=seed
            )
            X_posterior_1 = esmda_distance.assimilate_batch(
                X_batch=X_previous, Y=Y, rho_batch=rho
            )

            esmda_distance = DistanceESMDA(
                covariance=C_D, observations=true_observations, alpha=alpha_i, seed=seed
            )
            X_posterior_2 = esmda_distance.assimilate(X=X_previous, Y=Y, rho=rho)

            diff_X = X_posterior_2 - X_posterior_1
            max_diff = np.max(np.abs(diff_X))
            # Check that the results are equal
            assert max_diff < 1e-9, (
                "The function assimilate and assimilate_batch give different results"
            )

            esmda = ESMDA(
                covariance=C_D, observations=true_observations, alpha=alpha_i, seed=rng
            )

            X_posterior_global = esmda.assimilate(X=X_previous_global, Y=Y)
            assert_tests_for_distance_based_localization(
                X_prior,
                j_obs,
                rho,
                true_parameters,
                X_posterior_1,
                X_posterior_global,
                rho_min=1e-5,
                atol=1e-2,
            )


@pytest.mark.parametrize("seed", list(range(9)))
def test_distance_based_on_1D_case_with_rho_equal_one(seed):
    N_m = 100  # Number of model parameters (grid points)
    N_e = 50  # Ensemble size
    j_obs = 50  # Index of the single observation

    # Run the test over 9 different sets of alpha.
    nalpha = 10
    number_of_alpha_values = np.arange(1, nalpha, 1, dtype=np.int32)
    for m in number_of_alpha_values:
        alpha_vector = generate_alpha_vector(m)
        inverse_alpha_vector = 1.0 / alpha_vector

        # Consistency requirement for alpha
        # Sum of 1/alpha_vector[i] must be 1
        assert abs(inverse_alpha_vector.sum() - 1.0) < 1e-7

        obs_error_var = 0.01  # Variance of observation error
        true_observations = np.array([1.0])
        C_D = np.array([obs_error_var])
        for iteration in range(len(alpha_vector)):
            alpha_i = np.array([alpha_vector[iteration]])
            if iteration == 0:
                # Draw prior
                rng = np.random.default_rng(seed)
                X_prior = rng.normal(loc=0.0, scale=0.5, size=(N_m, N_e))
                X_previous = X_prior.copy()
                X_previous_global = X_prior.copy()
                X_posterior = X_prior
                X_posterior_global = X_prior
            else:
                X_previous = X_posterior
                X_previous_global = X_posterior_global
            # Predict observations `Y` using the identity model `g(x) = x`
            # We only observe the state at `j_obs`.
            Y = X_previous[[j_obs], :]

            # Set rho to 1 (no localization)
            rho = np.ones((N_m, 1), dtype=np.float64)
            # Must use same seed for both DistanceESMDA and ESMDA initialization
            # to be able to compare
            esmda_distance = DistanceESMDA(
                covariance=C_D, observations=true_observations, alpha=alpha_i, seed=seed
            )
            X_posterior = esmda_distance.assimilate(X=X_previous, Y=Y, rho=rho)

            esmda = ESMDA(
                covariance=C_D, observations=true_observations, alpha=alpha_i, seed=seed
            )
            X_posterior_global = esmda.assimilate(X=X_previous_global, Y=Y)

            diff_X = X_posterior - X_posterior_global
            max_diff = np.max(np.abs(diff_X))
            # Check that the results are equal
            assert max_diff < 1e-8, (
                "DistanceESMDA and ESMDA differs when using rho=1."
                f"Difference is: {max_diff}"
            )


@pytest.mark.parametrize(
    (
        "nparam",
        "nreal",
        "xinc",
        "corr_func_name",
        "relative_corr_range",
        "relative_localization_radius",
        "obs_err_std",
        "tolerance",
        "seed",
        "use_localization",
    ),
    [
        (
            2000,
            100,
            50.0,
            "exponential",
            0.05,
            2.5,
            0.001,
            0.17,
            13579,
            True,
        ),
        (
            2000,
            500,
            50.0,
            "exponential",
            0.05,
            2.5,
            0.001,
            0.11,
            987657,
            True,
        ),
        (
            2000,
            40000,
            50.0,
            "exponential",
            0.05,
            2.5,
            0.001,
            0.05,
            82872857,
            True,
        ),
        (
            2000,
            100,
            50.0,
            "exponential",
            0.05,
            2.5,
            0.001,
            0.17,
            13579,
            False,
        ),
        (
            2000,
            500,
            50.0,
            "exponential",
            0.05,
            2.5,
            0.001,
            0.11,
            987657,
            False,
        ),
        (
            2000,
            40000,
            50.0,
            "exponential",
            0.05,
            2.5,
            0.001,
            0.05,
            82872857,
            False,
        ),
    ],
)
def test_distance_based_localization_on_1D_corr_field(
    nparam: int,
    nreal: int,
    xinc: float,
    corr_func_name: str,
    relative_corr_range: float,
    relative_localization_radius: float,
    obs_err_std: float,
    tolerance: float,
    seed: int,
    use_localization: bool,
) -> None:
    # This test will check that Distance based ESMDA of the posterior mean
    # converge to a value close to the theoretical limit for an experiment
    # where observation errors are small and there are no trends in the
    # prior gaussian field. The theorietical limit when using ESMDA
    # should be exactly the same as the simple kriging estimate,
    # but the theoretical limit when using Distancebased localization
    # will depend on the ratio between the spatial correlation length
    # of the prior field and the localization range.
    # The expected result for distance based localization with ESMDA
    # when number of realizations approach a very large number
    # will correspond to gaussian field with a prior with shorter
    # correlation length than what is specified due to the downscaling
    # of the effect of the observations with distance. (Reduced value
    # of Kalman Gain matrix elements)

    # Prior field mean and stdev
    mean = 0.0
    stdev = 1.0

    corr_range = relative_corr_range * nparam * xinc
    obs_vector = np.array([5.0, -5.0, 4.0, 5.0], dtype=np.float64)
    obs_index_vector = np.array(
        [
            int(nparam / 10),
            int(3 * nparam / 10),
            int(6 * nparam / 10),
            int(9 * nparam / 10),
        ],
        dtype=np.int32,
    )
    nobs = obs_index_vector.shape[0]
    alpha = np.array([1.0])
    obs_error_var = obs_err_std**2  # Variance of observation error
    C_D = np.array([obs_error_var] * nobs)

    # Draw prior ensemble with specified spatial correlation function
    rng = np.random.default_rng(seed)
    X_prior, field_cov_matrix = draw_1D_field(
        mean, stdev, xinc, corr_func_name, corr_range, nparam, nreal, rng
    )

    # Calculate theoretical posterior mean
    # assuming prior mean equal 0.
    # Assume forward model is identity  X = g(X)
    # Use simple kriging estimate for theoretical posterior mean
    mean_predicted_field = predicted_mean_field_values(
        obs_vector, obs_index_vector, field_cov_matrix
    )

    # Predict observations `Y` using the identity model `g(x) = x`
    Y = X_prior[obs_index_vector, :]

    # Using a simple Gaussian decay for rho.
    localization_radius = relative_localization_radius * corr_range

    rho_matrix = np.zeros((nparam, nobs), dtype=np.float64)
    if use_localization:
        for i in range(nobs):
            rho = calculate_rho_1d(
                nparam,
                obs_index_vector[i],
                localization_radius,
                xinc=xinc,
            )
            rho_matrix[:, i] = rho[:, 0]
    else:
        rho_matrix[:, :] = 1.0

    esmda_distance = DistanceESMDA(
        covariance=C_D, observations=obs_vector, alpha=alpha, seed=seed
    )
    X_posterior = esmda_distance.assimilate(X=X_prior, Y=Y, rho=rho_matrix)

    esmda = ESMDA(covariance=C_D, observations=obs_vector, alpha=alpha, seed=seed)
    X_posterior_global = esmda.assimilate(X=X_prior, Y=Y)

    # Mean and stdev of ensemble of posterior field
    X_post_mean = X_posterior.mean(axis=1)
    X_post_mean_global = X_posterior_global.mean(axis=1)

    # Difference with theoretical mean based on simple kriging is calculated.
    # Note that ESMDA in this test should approach the kriging estimate
    # which is the theoretical limit of the posterior mean when nreal -> infinity
    # For DL-ESMDA only when localization range is much larger than the spatial
    # correlation length one can expect the posterior mean approaches the
    # simple kriging estimate in this test. But DL-ESMDA will be closer to
    # the simple kriging estimate for practical purposes where nreal is not large.
    # The tolerance specified in this test is specified such that the estimated
    # standard deviation of the differences between DL-ESMDA and simple kriging
    # estimate is less than the tolerances. Note that when nreal increases, the
    # tolerances can be reduced, but will never approach 0 as long as
    # localization range is finite. However, for ESMDA, the tolerance
    # could in theory approach 0 when nreal approach infinity
    X_diff_local_sk = X_post_mean - mean_predicted_field
    X_diff_global_sk = X_post_mean_global - mean_predicted_field
    X_diff_local_abs = np.abs(X_diff_local_sk)
    X_diff_global_abs = np.abs(X_diff_global_sk)
    est_std_diff_global = X_diff_global_abs.std()
    est_std_diff_local = X_diff_local_abs.std()
    X_local_diff = np.max(X_diff_local_abs)
    X_global_diff = np.max(X_diff_global_abs)

    print(f"Number of real: {nreal}")
    print(f"Max difference using DL ESMDA : {X_local_diff}")
    print(f"Max difference using ordinary ESMDA: {X_global_diff}")
    print(
        "Estimated std of difference between DL and simple kriging: "
        f"{est_std_diff_local}"
    )
    print(
        "Estimated std of difference between Global and simple kriging: "
        f"{est_std_diff_global}"
    )
    if use_localization:
        assert est_std_diff_local <= tolerance, (
            "Estimated std of difference between DL "
            f"and simple kriging estimate is {est_std_diff_local} "
            f"is above specified tolerance {tolerance}"
        )
    else:
        # In this case both ESMDA and DL-ESMDA should give the same result
        # since there is no localization (rho=1 for every element in rho)
        assert abs(est_std_diff_global - est_std_diff_local) < 1e-7


@pytest.mark.parametrize(
    ("nx", "ny", "nz", "num_obs", "nreal", "bytes_per_float"),
    [
        (
            100,
            100,
            50,
            100,
            100,
            8,
        ),
        (
            100,
            100,
            100,
            1000,
            100,
            4,
        ),
        (
            750,
            650,
            500,
            1000,
            100,
            4,
        ),
        (
            100,
            100,
            100,
            10000,
            100,
            4,
        ),
        (
            100,
            100,
            100,
            10000,
            500,
            8,
        ),
        (
            250,
            350,
            500,
            1000,
            500,
            4,
        ),
    ],
)
def test_calc_max_number_of_layers_per_batch_for_distance_localization(
    nx: int,
    ny: int,
    nz: int,
    num_obs: int,
    nreal: int,
    bytes_per_float: int,
) -> None:
    # Test for function that estimate max number of grid layers with field parameters
    # that can be update in one update. No expected number of layers can be defined
    # in advance since this depends on the available memory of the computer that run
    # the test.
    max_nlayer_per_batch = (
        calc_max_number_of_layers_per_batch_for_distance_localization(
            nx, ny, nz, num_obs, nreal, bytes_per_float=bytes_per_float
        )
    )

    memory_used = max_nlayer_per_batch * nx * ny * num_obs * 2 * bytes_per_float / 10**9
    if max_nlayer_per_batch == nz:
        print(f"\nAll {nz} grid layers can be updated in one batch.")
    else:
        print(
            f"\nMax number of layers per batch: {max_nlayer_per_batch}. "
            f"Total number of layers: {nz}"
        )
    print(f"Memory per batch used for RHO and K matrix: {memory_used} GB")


@pytest.mark.parametrize(
    ("nx", "ny", "nz", "nobs", "nreal", "min_nbatch"),
    [
        (
            10,
            15,
            100,
            10,
            100,
            4,
        ),
        (
            10,
            15,
            100,
            10,
            10,
            1,
        ),
        (
            100,
            150,
            1,
            10,
            100,
            1,
        ),
        (
            100,
            150,
            1,
            0,
            100,
            1,
        ),
        # (
        #     412,
        #     714,
        #     51,
        #     1500,
        #     100,
        #     1,
        # ),
    ],
)
def test_batch_handling_for_update_params_3D(
    nx: int,
    ny: int,
    nz: int,
    nobs: int,
    nreal: int,
    min_nbatch: int,
):
    # This test function checks only the batch split part of the
    # member function 'update_params_3D' of the class DistanceESMDA
    # The prior ensemble will only be copied, there is no update
    # The purpose is to check that the field parameter values come
    # in correct order (no index error)

    nparam = nx * ny * nz
    # Define a prior ensemble where each field parameter value for each realization
    # is unique. Then it is easy to check that indexing is correct by comparing
    # input with output
    X_prior = np.arange(nparam * nreal, dtype=np.float64).reshape(nparam, nreal)
    observations = np.zeros(nobs, dtype=np.float64)
    obs_var_vector = np.zeros(nobs, dtype=np.float64)
    Y = np.zeros((nobs, nreal), dtype=np.float64)
    rho_2D = np.zeros((nx, ny, nobs), dtype=np.float64)

    # Initialize Distance based localization object using the mock version
    # that does not change any values
    alpha = np.array([1.0])
    dl_smoother = DistanceESMDA(
        covariance=obs_var_vector, observations=observations, alpha=alpha, seed=42
    )

    # No update will be done here and the only thing to happen
    # is that input X_prior is returned
    X_post = dl_smoother.update_params(
        X_prior,
        Y,
        rho_2D,
        nz,
        min_nbatch=min_nbatch,
        no_update=True,
    )

    # Check that X_prior = X_post
    assert X_prior.shape == X_post.shape
    assert_allclose(X_prior, X_post)


def draw_3D_field(
    mean: float,
    stdev: float,
    xinc: float,
    yinc: float,
    zinc: float,
    nx: int,
    ny: int,
    nz: int,
    nreal: int,
    main_corr_range: float,
    perp_corr_range: float,
    vert_corr_range: float,
    start_seed: int = 42,
    corr_func_name: str = "matern32",
    power: float = 1.9,
    azimuth: float = 0.0,
    dip: float = 0.0,
    write_progress: bool = False,
    use_4_byte_float: bool = False,
) -> npt.NDArray[np.float64]:
    # Draw prior ensemble of 3D field. Ensemble size is nreal
    # Returns prior ensemble drawn and covariance matrix

    # Initialize start seed
    import gaussianfft as grf  # noqa

    grf.seed(start_seed)

    # Define spatial correlation function for gaussian fields
    # to be simulated.
    if corr_func_name == "general_exponential":
        variogram = grf.variogram(
            corr_func_name,
            main_corr_range,
            perp_corr_range,
            vert_corr_range,
            azimuth,
            dip,
            power,
        )
    else:
        variogram = grf.variogram(
            corr_func_name,
            main_corr_range,
            perp_corr_range,
            vert_corr_range,
            azimuth,
            dip,
        )

    nparam = nx * ny * nz
    if use_4_byte_float:
        X_prior = np.zeros((nparam, nreal), dtype=np.float32)
    else:
        X_prior = np.zeros((nparam, nreal), dtype=np.float64)
    # Flatten 3D array in F order
    for real_number in range(nreal):
        if write_progress and real_number % 10 == 0:
            print(f"  Sim real nr: {real_number}")
        field_values = grf.simulate(variogram, nx, xinc, ny, yinc, nz, zinc)
        X_prior[:, real_number] = (
            field_values.reshape((nx, ny, nz), order="F").flatten(order="C") * stdev
            + mean
        )

    return X_prior


def draw_random_obs(rng, nobs, nx, ny, nz, obs_err_std):
    nparam = nx * ny * nz
    assert nobs < nparam
    # Define a grid resolution with specified size
    xinc = 50.0
    yinc = 50.0
    zinc = 1.0
    right_handed_grid_indexing = True
    # Draw some observation values (Use same seed every time)
    observations = rng.normal(loc=0.5, scale=0.05, size=nobs)

    # Choose some observation values, errors and positions
    # Draw some position of the observations, ensure no observations at same position
    # since the response values are equal to field values (only one response variable).
    # Multiple response values can have same position.

    # Draw i index, j_index, k_index for grid cell to be used as observed.
    if nobs == 1:
        obs_xpos = np.array([(nx / 2) * xinc])
        obs_ypos = np.array([(ny / 2) * yinc])
        obs_zpos = np.array([(nz / 2) * zinc])
        i_indices = np.array([int(nx / 2)])
        j_indices = np.array([int(ny / 2)])
        k_indices = np.array([int(nz / 2)])
        unique_obs_indices = k_indices + j_indices * nz + i_indices * nz * ny
    else:
        if nobs > nparam:
            raise ValueError(
                "Cannot draw more observations than number of parameters in the field"
            )
        unique_obs_indices = rng.choice(range(nparam), size=nobs, replace=False)
        unique_obs_indices = np.sort(unique_obs_indices)
        i_indices = (unique_obs_indices // (nz * ny)).astype(int)
        j_indices = ((unique_obs_indices % (nz * ny)) // nz).astype(int)
        k_indices = (unique_obs_indices % nz).astype(int)
        if right_handed_grid_indexing:
            # Right-handed grid indexing
            obs_ypos = ((ny - j_indices - 1) + 0.5) * yinc
        else:
            obs_ypos = (j_indices + 0.5) * yinc

        obs_xpos = (i_indices + 0.5) * xinc
        obs_zpos = (k_indices + 0.5) * zinc

    # Set observation error
    obs_var_vector = np.zeros(nobs, dtype=np.float64)
    obs_var_vector[:] = obs_err_std**2
    return (
        observations,
        obs_var_vector,
        obs_xpos,
        obs_ypos,
        obs_zpos,
        i_indices,
        j_indices,
        k_indices,
        unique_obs_indices,
    )


@pytest.mark.skipif(
    sys.version_info >= (3, 13),
    reason="gaussianfft not available on Python >= 3.13",
)
@pytest.mark.parametrize(
    (
        "nx",
        "ny",
        "nz",
        "nobs",
        "nreal",
        "field_mean",
        "field_std",
        "rel_corr_length",
        "obs_err_std",
        "rel_localization_range",
        "case_with_some_zero_variance_field_params",
        "seed",
        "name",
    ),
    [
        (
            3,
            4,
            5,
            1,
            100,
            0.0,
            1.0,
            0.2,
            0.001,
            0.2,
            False,
            9984356,
            "case1",
        ),
        (
            5,
            4,
            3,
            0,
            100,
            0.0,
            1.0,
            0.1,
            0.01,
            0.1,
            False,
            123456,
            "case2",
        ),
        (
            5,
            4,
            3,
            3,
            100,
            0.0,
            1.0,
            0.1,
            0.01,
            0.1,
            True,
            8765,
            "case3",
        ),
        (
            3,
            5,
            1,
            1,
            100,
            0.0,
            1.0,
            0.3,
            0.001,
            0.3,
            False,
            9984356,
            "case4",
        ),
        (
            3,
            5,
            1,
            1,
            100,
            0.0,
            1.0,
            0.3,
            0.001,
            0.3,
            True,
            9984356,
            "case5",
        ),
    ],
)
def test_update_params_3D(
    snapshot,
    nx: int,
    ny: int,
    nz: int,
    nobs: int,
    nreal: int,
    field_mean: float,
    field_std: float,
    rel_corr_length: float,
    obs_err_std: float,
    rel_localization_range: float,
    case_with_some_zero_variance_field_params: bool,
    seed: int,
    name: str,
):
    # This test function check 'update_3D_field_with_distance_esmda_unoptimal'
    #
    # The field parameter is assumed to belong to a box grid (ertbox grid)
    # with specified nx,ny,nz and grid increments xinc,yinc,zinc
    # The observation position is assumed to be within the same coordinate
    # system as the grid. The grid cell center point coordinates are
    # x[i,j,k] = xinc * (i + 0.5)  i=0,.. nx-1
    # y[i,j,k] = yinc * (j + 0.5)  j=0,.. ny-1
    # z[i,j,k] = zinc * (k + 0.5)  k=0,.. nz-1
    # Z coordinate is not used when calculating RHO matrix, but is used here
    # to define observation values.
    right_handed_grid_indexing = True  # Default is True
    write_obs = False
    xinc = 50.0
    yinc = 50.0
    zinc = 1.0
    xlength = xinc * nx
    ylength = yinc * ny
    zlength = zinc * nz
    corr_range = max(xlength, ylength) * rel_corr_length
    vert_range = zlength * rel_corr_length
    fraction_of_field_values_with_zero_variance = 0.1
    # Draw prior gaussian fields with spatial correlations
    # X_prior[param_number,real_number] where param_number
    # is flatten indec for C-index ordered 3D field (nx,ny,nz)
    corr_func_name = "gaussian"
    X_prior = draw_3D_field(
        field_mean,
        field_std,
        xinc,
        yinc,
        zinc,
        nx,
        ny,
        nz,
        nreal,
        corr_range,
        corr_range,
        vert_range,
        seed,
        corr_func_name=corr_func_name,
    )

    X_prior_3D = X_prior.reshape((nx, ny, nz, nreal))
    rng = np.random.default_rng(seed)
    nparam = nx * ny * nz
    (
        observations,
        obs_var_vector,
        obs_xpos,
        obs_ypos,
        obs_zpos,
        i_indices,
        j_indices,
        k_indices,
        unique_obs_indices,
    ) = draw_random_obs(rng, nobs, nx, ny, nz, obs_err_std)

    # Choose localization range around each obs
    typical_field_size = min(xlength, ylength)
    obs_main_range = np.zeros(nobs, dtype=np.float64)
    obs_main_range[:] = typical_field_size * rel_localization_range
    obs_perp_range = np.zeros(nobs, dtype=np.float64)
    obs_perp_range[:] = typical_field_size * rel_localization_range
    obs_anisotropy_angle = np.zeros(nobs, dtype=np.float64)

    if write_obs:
        print("Observations:")
        for i in range(nobs):
            print(
                f"({obs_xpos[i]}, {obs_ypos[i]}, {obs_zpos[i]})  "
                f"Value: {observations[i]}  Error: {obs_err_std}  "
                f"Range: {obs_main_range[i]}"
            )

    # Calculate rho_for one layer
    rho_2D = calc_rho_for_2d_grid_layer(
        nx,
        ny,
        xinc,
        yinc,
        obs_xpos,
        obs_ypos,
        obs_main_range,
        obs_perp_range,
        obs_anisotropy_angle,
        right_handed_grid_indexing=right_handed_grid_indexing,
    )
    # Set responses for each observation equal to the X_prior for simplicity
    # (Forward model is identity Y = X in observation points + small random noise)
    # Note cannot have observations with response with 0 variance
    add_response_variability = rng.normal(loc=0, scale=0.01, size=(nreal, nobs))
    Y = X_prior_3D[i_indices, j_indices, k_indices, :] + add_response_variability.T

    if case_with_some_zero_variance_field_params:
        # Set same field value in all realizations for selected grid cells
        # They should not be updated since the ensemble variance is 0 for those
        # values. The selected grid cells must not be selected as observed.
        nconst_values = int(
            fraction_of_field_values_with_zero_variance * nparam
        )  # Choose a portion of field values to be constant
        unique_indices_const = rng.choice(
            range(nparam), size=nconst_values, replace=False
        )
        print(f"Number of selected grid indices: {len(unique_indices_const)}")
        # Reject indices corresponding to observed grid cells
        unique_indices_const = np.setdiff1d(unique_indices_const, unique_obs_indices)
        # Usually inactive cell values can be set to 0 for all realizations,
        # but choose something else here for visualization purpose
        X_prior[unique_indices_const, :] = 10.0
        print(f"Number of field values with 0 variance: {len(unique_indices_const)}")

    # Initialize Distance based localization object
    alpha = np.array([1.0])
    dl_smoother = DistanceESMDA(
        covariance=obs_var_vector, observations=observations, alpha=alpha, seed=rng
    )

    # Call the function to be tested here
    # Note that no field values with 0 variance is removed
    # from the update calculation in this function,
    # but that is ok since there will be no change of
    # field parameters with 0 variance anyway.
    X_post = dl_smoother.update_params(
        X_prior,
        Y,
        rho_2D,
        nz,
    )
    X_post_3D = X_post.reshape((nx, ny, nz, nreal))
    # Mean and stdev over ensemble of field parameters
    #    X_prior_mean_3D = X_prior_3D.mean(axis=3)
    X_post_mean_3D = X_post_3D.mean(axis=3)
    #    X_prior_stdev_3D = X_prior_3D.std(axis=3)
    X_post_stdev_3D = X_post_3D.std(axis=3)
    # Difference between post and prior mean and stdev
    #    X_diff_mean_3D = X_post_mean_3D - X_prior_mean_3D
    #    X_diff_stdev_3D = X_post_stdev_3D - X_prior_stdev_3D

    # Check results
    #    snapshot.assert_match(str(X_diff_mean_3D) + "\n", "X_diff_mean_3D.txt")
    #    snapshot.assert_match(str(X_diff_stdev_3D) + "\n", "X_diff_stdev_3D.txt")
    snapshot.assert_match(str(X_post_mean_3D) + "\n", "X_post_mean_3D.txt")
    snapshot.assert_match(str(X_post_stdev_3D) + "\n", "X_post_stdev_3D.txt")


if __name__ == "__main__":
    pytest.main(
        args=[
            __file__,
            "-v",
        ]
    )



"""
Contains (publicly available, but not officially supported) experimental
features of iterative_ensemble_smoother
"""

import logging
import numbers
import warnings
from typing import Callable, List, Optional, Sequence, Tuple, TypeVar, Union

import numpy as np
import numpy.typing as npt
import scipy as sp  # type: ignore
from joblib import Parallel, delayed

from iterative_ensemble_smoother import ESMDA
from iterative_ensemble_smoother.esmda import BaseESMDA
from iterative_ensemble_smoother.esmda_inversion import (
    empirical_cross_covariance,
    singular_values_to_keep,
)
from iterative_ensemble_smoother.utils import (
    calc_max_number_of_layers_per_batch_for_distance_localization,
    memory_usage_decorator,
)

logger = logging.getLogger(__name__)

T = TypeVar("T")

ENABLE_MEMORY_LOGGING_FOR_DISTANCE_BASED_LOCALIZATION = True


class AdaptiveESMDA(BaseESMDA):
    @staticmethod
    def correlation_threshold(ensemble_size: int) -> float:
        """Return a number that determines whether a correlation is significant.

        Default threshold taken from luo2022,
        Continuous Hyper-parameter OPtimization (CHOP) in an ensemble Kalman filter
        Section 2.3 - Localization in the CHOP problem

        Examples
        --------
        >>> AdaptiveESMDA.correlation_threshold(0)
        1.0
        >>> AdaptiveESMDA.correlation_threshold(9)
        1.0
        >>> AdaptiveESMDA.correlation_threshold(16)
        0.75
        >>> AdaptiveESMDA.correlation_threshold(36)
        0.5
        """
        return float(min(1, max(0, 3 / np.sqrt(ensemble_size))))

    @staticmethod
    def compute_cross_covariance_multiplier(
        *,
        alpha: float,
        C_D: npt.NDArray[np.double],
        D: npt.NDArray[np.double],
        Y: npt.NDArray[np.double],
        cov_YY: Optional[npt.NDArray[np.double]] = None,
    ) -> npt.NDArray[np.double]:
        """Compute transition matrix T such that:

            X + cov_XY @ transition_matrix

        In the notation of Emerick et al, recall that the update equation is:

            X_posterior = X_prior + cov_XY @ inv(C_DD + alpha * C_D) @ (D - Y)

        This function computes a transition matrix T, defined as:

            T := inv(C_DD + alpha * C_D) @ (D - Y)

        Note that this might not be the most efficient way to compute the update,
        since cov_XY := center(X) @ center(Y) / (N - 1), and we can avoid creating
        this (huge) covariance matrix by performing multiplications right-to-left:

            cov_XY @ inv(C_DD + alpha * C_D) @ (D - Y)
            center(X) @ center(Y) / (N - 1) @ inv(C_DD + alpha * C_D) @ (D - Y)
            center(X) @ [center(Y) / (N - 1) @ inv(C_DD + alpha * C_D) @ (D - Y)]

        The advantage of forming T instead is that if we already have computed
        cov_XY, then we can apply the same T to a reduced number of rows (parameters)
        in cov_XY.
        """
        # Compute cov(Y, Y) if it was not passed to the function.
        # Pre-computation might be faster, since covariance is commutative with
        # respect to indexing, ie, cov(Y[mask, :], YY[mask, :]) = cov(Y, Y)[mask, mask]
        C_DD = empirical_cross_covariance(Y, Y) if cov_YY is None else cov_YY

        assert C_DD.shape[0] == C_DD.shape[1]
        assert C_DD.shape[0] == Y.shape[0]

        # Arguments for sp.linalg.solve
        solver_kwargs = {
            "overwrite_a": True,
            "overwrite_b": True,
            "assume_a": "pos",  # Assume positive definite matrix (use cholesky)
            "lower": False,  # Only use the upper part while solving
        }

        # Compute T := sp.linalg.inv(C_DD + alpha * C_D) @ (D - Y)
        # by solving the system (C_DD + alpha * C_D) @ T = (D - Y)
        if C_D.ndim == 2:
            # C_D is a covariance matrix
            C_DD += alpha * C_D  # Save memory by mutating
        elif C_D.ndim == 1:
            # C_D is an array, so add it to the diagonal without forming diag(C_D)
            np.fill_diagonal(C_DD, C_DD.diagonal() + alpha * C_D)

        return sp.linalg.solve(C_DD, D - Y, **solver_kwargs)  # type: ignore

    def _cov_to_corr_inplace(
        self,
        cov_XY: npt.NDArray[np.double],
        stds_X: npt.NDArray[np.double],
        stds_Y: npt.NDArray[np.double],
    ) -> None:
        """Convert a covariance matrix to a correlation matrix in-place."""

        # Divide each element of cov_XY by the corresponding standard deviations
        cov_XY /= stds_X[:, np.newaxis]
        cov_XY /= stds_Y[np.newaxis, :]

        # Perform checks and clip values to [-1, 1]
        eps = 1e-8
        if not ((cov_XY.max() <= 1 + eps) and (cov_XY.min() >= -1 - eps)):
            msg = "Cross-correlation matrix has entries not in [-1, 1]."
            msg += f"The min and max values are: {cov_XY.min()} and {cov_XY.max()}"
            warnings.warn(msg)

        return np.clip(cov_XY, a_min=-1, a_max=1, out=cov_XY)

    def _corr_to_cov_inplace(
        self,
        corr_XY: npt.NDArray[np.double],
        stds_X: npt.NDArray[np.double],
        stds_Y: npt.NDArray[np.double],
    ) -> None:
        """Convert a correlation matrix to a covariance matrix in-place."""
        # Multiply each element of corr_XY by the corresponding standard deviations
        corr_XY *= stds_X[:, np.newaxis]
        corr_XY *= stds_Y[np.newaxis, :]
        return corr_XY

    @staticmethod
    def _update_single_parameter(
        param_num: int,
        correlated_responses_mask: npt.NDArray[np.bool_],
        cov_XY_row: npt.NDArray[np.double],
        Y: npt.NDArray[np.double],
        D: npt.NDArray[np.double],
        C_D: npt.NDArray[np.double],
        cov_YY: npt.NDArray[np.double],
        alpha: float,
        compute_cross_covariance_multiplier: Callable,
    ) -> Tuple[int, npt.NDArray[np.double]]:
        """
        Worker function to compute the update for a single parameter.
        """

        cov_XY_subset = cov_XY_row[correlated_responses_mask].reshape(1, -1)

        Y_subset = Y[correlated_responses_mask, :]
        D_subset = D[correlated_responses_mask, :]

        cov_YY_mask = np.ix_(correlated_responses_mask, correlated_responses_mask)
        cov_YY_subset = cov_YY[cov_YY_mask]

        C_D_subset = (
            C_D[correlated_responses_mask] if C_D.ndim == 1 else C_D[cov_YY_mask]
        )

        T = compute_cross_covariance_multiplier(
            alpha=alpha,
            C_D=C_D_subset,
            D=D_subset,
            Y=Y_subset,
            cov_YY=cov_YY_subset,
        )

        update_vector = cov_XY_subset @ T

        return param_num, update_vector

    def assimilate(
        self,
        *,
        X: npt.NDArray[np.double],
        Y: npt.NDArray[np.double],
        D: npt.NDArray[np.double],
        overwrite: bool = False,
        alpha: float,
        correlation_threshold: Union[Callable[[int], float], float, None] = None,
        cov_YY: Optional[npt.NDArray[np.double]] = None,
        progress_callback: Optional[Callable[[Sequence[T]], Sequence[T]]] = None,
        correlation_callback: Optional[Callable[[npt.NDArray[np.double]], None]] = None,
        n_jobs: int = 1,
    ) -> npt.NDArray[np.double]:
        """Assimilate data and return an updated ensemble X_posterior.

            X_posterior = smoother.assimilate(X, Y, D, alpha)

        This method first computes the cross-covariance and cross-correlation
        matrices between X and Y. Then it sets correlations that are below the
        threshold to zero. It then loops over parameters in X that are deemed
        significant with respect to Y (based on the threshold), and updates
        these groups together.

        Parameters
        ----------
        X : np.ndarray
            A 2D array of shape (num_parameters, ensemble_size). Each row
            corresponds to a parameter in the model, and each column corresponds
            to an ensemble member (realization).
        Y : np.ndarray
            2D array of shape (num_observations, ensemble_size), containing
            responses when evaluating the model at X. In other words, Y = g(X),
            where g is the forward model.
        D : np.ndarray
            2D array of shape (num_observations, ensemble_size), containing
            perturbed observations. D = observations + mv_normal(0, covariance),
            and D may be computed by the method `perturb_observations`.
        alpha : float
            The covariance inflation factor. The sequence of alphas should
            obey the equation sum_i (1/alpha_i) = 1. However, this is NOT
            enforced in this method. The user/caller is responsible for this.
        overwrite: bool
            If True, X will be overwritten and mutated.
            If False, the method will not mutate inputs in any way.
            Setting this to True saves memory.
        correlation_threshold : callable or float or None
            Either a callable with signature f(ensemble_size) -> float, or a
            float in the range [0, 1]. Entries in the covariance matrix that
            are lower than the correlation threshold will be set to zero.
            If None, the default 3/sqrt(ensemble_size) is used.
        cov_YY : np.ndarray or None
            A 2D array of shape (num_observations, num_observations) with the
            empirical covariance of Y. If passed, this is not computed in the
            method call, potentially saving time and computation.
        progress_callback : Callable[[Sequence[T]], Sequence[T]] or None
            A callback function that can be used to wrap the iteration over
            parameters for progress reporting.
            It should accept an iterable as input and return an iterable.
            This allows for integration with progress reporting tools like tqdm,
            which can provide visual feedback on the progress of the
            assimilation process.
            If None, no progress reporting is performed.
        correlation_callback : Optional[Callable]
            A callback function that is called with the correlation matrix (2D array)
            as its argument after the correlation matrix computation is complete.
            The callback should handle or process the correlation matrix, such as
            saving or logging it. The callback should not return any value.
        n_jobs : int
            The number of parallel jobs to run. If 1, no parallel processing
            is used. If -1, all available CPU cores are used. Default is 1.

        Returns
        -------
        X_posterior : np.ndarray
            2D array of shape (num_parameters, ensemble_size).
        """
        assert X.shape[1] == Y.shape[1]

        # Check the correlation threshold
        is_callable = callable(correlation_threshold)
        is_float = (
            isinstance(correlation_threshold, numbers.Real)
            and correlation_threshold >= 0
            and correlation_threshold <= 1
        )
        is_None = correlation_threshold is None
        if not (is_callable or is_float or is_None):
            raise TypeError(
                "`correlation_threshold` must be a callable or a float in [0, 1]"
            )

        # Do not overwrite input arguments
        if not overwrite:
            X = np.copy(X)

        # Create `correlation_threshold` if the argument is a float
        if is_float:
            corr_threshold: float = correlation_threshold

            def correlation_threshold(ensemble_size: int) -> float:
                return corr_threshold

        # Default correlation threshold function
        if correlation_threshold is None:
            correlation_threshold = self.correlation_threshold
        assert callable(correlation_threshold), (
            "`correlation_threshold` should be callable"
        )

        if progress_callback is None:

            def progress_callback(x):
                return x  # A simple pass-through function

        # Step 1: # Compute cross-correlation between parameters X and responses Y
        # Note: let the number of parameters be n and the number of responses be m.
        # This step requires both O(mn) computation and O(mn) storage, which is
        # larger than the O(n + m^2) computation used in ESMDA, which never explicitly
        # forms the cross-covariance matrix between X and Y. However, if we batch
        # X by parameter group (rows), the storage requirement decreases
        cov_XY = empirical_cross_covariance(X, Y)
        assert cov_XY.shape == (X.shape[0], Y.shape[0])

        stds_X = np.std(X, axis=1, ddof=1)
        stds_Y = np.std(Y, axis=1, ddof=1)

        corr_XY = self._cov_to_corr_inplace(cov_XY, stds_X, stds_Y)

        # Determine which elements in the cross covariance matrix that will
        # be set to zero
        threshold = correlation_threshold(X.shape[1])
        significant_corr_XY = np.abs(corr_XY) > threshold

        cov_XY = self._corr_to_cov_inplace(corr_XY, stds_X, stds_Y)

        # Pre-compute the covariance cov(Y, Y) here, and index on it later
        if cov_YY is None:
            cov_YY = empirical_cross_covariance(Y, Y)
        else:
            assert cov_YY.ndim == 2, "'cov_YY' must be a 2D array"
            assert cov_YY.shape == (Y.shape[0], Y.shape[0])

        # Identify rows with at least one significant correlation.
        significant_rows = np.any(significant_corr_XY, axis=1)

        # Loop only over rows with significant correlations
        params_to_update = np.where(significant_rows)[0]

        results = Parallel(n_jobs=n_jobs)(
            delayed(self._update_single_parameter)(
                param_num,
                significant_corr_XY[param_num],
                cov_XY[param_num],
                # These are sent in full, which is efficient as they are the same
                # for all workers. Joblib handles this well.
                Y,
                D,
                self.C_D,
                cov_YY,
                alpha,
                self.compute_cross_covariance_multiplier,
            )
            for param_num in progress_callback(params_to_update)
        )

        for param_num, update_vector in results:
            X[[param_num], :] += update_vector

        if correlation_callback is not None:
            corr_XY = self._cov_to_corr_inplace(cov_XY, stds_X, stds_Y)
            correlation_callback(corr_XY)

        return X


class RowScaling:
    # Illustration of how row scaling works, `multiply` is the important part
    # For the actual implementation, which is more involved, see:
    # https://github.com/equinor/ert/blob/84aad3c56e0e52be27b9966322d93dbb85024c1c/src/clib/lib/enkf/row_scaling.cpp
    def __init__(self, alpha=1.0):
        """Alpha is the strength of the update."""
        assert 0 <= alpha <= 1.0
        self.alpha = alpha

    def multiply(self, X, K):
        """Takes a matrix X and a matrix K and performs alpha * X @ K."""
        # This implementation merely mimics how RowScaling::multiply behaves
        # in the C++ code. It mutates the input argument X instead of returning.
        X[:, :] = X @ (K * self.alpha)


def ensemble_smoother_update_step_row_scaling(
    *,
    covariance: npt.NDArray[np.double],
    observations: npt.NDArray[np.double],
    X_with_row_scaling: List[Tuple[npt.NDArray[np.double], RowScaling]],
    Y: npt.NDArray[np.double],
    seed: Union[np.random._generator.Generator, int, None] = None,
    inversion: str = "exact",
    truncation: float = 1.0,
):
    """Perform a single ESMDA update (ES) with row scaling.
    The matrices in X_with_row_scaling WILL BE MUTATED.
    See the ESMDA class for information about input arguments.


    Explanation of row scaling
    --------------------------

    The ESMDA update can be written as:

        X_post = X_prior + X_prior @ K

    where K is a transition matrix. The core of the row scaling approach is that
    for each row i in the matrix X, we apply an update with strength alpha:

        X_post = X_prior + alpha * X_prior @ K
        X_post = X_prior @ (I + alpha * K)

    Clearly 0 <= alpha <= 1 denotes the 'strength' of the update; alpha == 1
    corresponds to a normal smoother update and alpha == 0 corresponds to no
    update. With the per row transformation of X the operation is no longer matrix
    multiplication but the pseudo code looks like:

        for i in rows:
            X_i_post = X_i_prior @ (I + alpha_i * K)

    See also original code:
        https://github.com/equinor/ert/blob/84aad3c56e0e52be27b9966322d93dbb85024c1c/src/clib/lib/enkf/row_scaling.cpp#L51

    """

    # Create ESMDA instance and set alpha=1 => run single assimilation (ES)
    smoother = ESMDA(
        covariance=covariance,
        observations=observations,
        seed=seed,
        inversion=inversion,
        alpha=1,
    )

    # Create transition matrix - common to all parameters in X
    transition_matrix = smoother.compute_transition_matrix(
        Y=Y, alpha=1, truncation=truncation
    )

    # The transition matrix K is a matrix such that
    #     X_posterior = X_prior + X_prior @ K
    # but the C++ code in ERT requires a transition matrix F that obeys
    #     X_posterior = X_prior @ F
    # To accomplish this, we add the identity to the transition matrix in place
    np.fill_diagonal(transition_matrix, transition_matrix.diagonal() + 1)

    # Loop over groups of rows (parameters)
    for X, row_scale in X_with_row_scaling:
        # In the C++ code, multiply() will transform the transition matrix F as
        #    F_new = F * alpha + I * (1 - alpha)
        # but the transition matrix F that we pass below is F := K + I, so:
        #    F_new = (K + I) * alpha + I * (1 - alpha)
        #    F_new = K * alpha + I * alpha + I - I * alpha
        #    F_new = K * alpha + I
        # And the update becomes : X_posterior = X_prior @ F_new
        # The update in the C++ code is equivalent to
        #    X_posterior = X_prior + alpha * X_prior @ K
        # if we had used the original transition matrix K that is returned from
        # ESMDA.compute_transition_matrix
        row_scale.multiply(X, transition_matrix)

    return X_with_row_scaling


class DistanceESMDA(ESMDA):
    def __init__(
        self,
        covariance: npt.NDArray[np.float64],
        observations: npt.NDArray[np.float64],
        alpha: Union[int, npt.NDArray[np.float64]] = 5,
        seed: Union[np.random._generator.Generator, int, None] = None,
    ) -> None:
        """
        Initialize instance.
        """
        # Initialize instance
        super().__init__(
            covariance=covariance, observations=observations, alpha=alpha, seed=seed
        )

        # Ensure self.X3 is initialized to None
        # Is set in prepare_assimilation and used in assimilate_batch
        # Is not used when using assimilate
        self.X3: npt.NDArray[np.float64] = None

    @memory_usage_decorator(
        enabled=ENABLE_MEMORY_LOGGING_FOR_DISTANCE_BASED_LOCALIZATION
    )
    def assimilate(
        self,
        *,
        X: npt.NDArray[np.float64],
        Y: npt.NDArray[np.float64],
        rho: npt.NDArray[np.float64],
        truncation: float = 0.99,
    ):
        """
        Implementation of algorithm described in Appendix B of Emerick's publication.
        Reference: Emerick, Journal of Petroleum Science and Engineering 136(3):219-239
                   DOI:10.1016/j.petrol.2016.01.029  (2016)

        Calculate Ensemble Smoother update with distance-based localization using the
        RHO matrix to define the scaling/tapering function for reduction of observations
        influence on field parameters. This function implements all steps in
        the algorithm.

        Typical workflow:
          Create object of DistanceESMDA
          Loop over batches of field parameter values from a field parameter:
            Run 'assimilate' for each batch of field parameter values

        Args:
            X: Parameter matrix shape=(nparameters, nrealizations)
            Y: Response matrix with predictions of observations,
               shape=(nobservations, nrealizations)
            rho: Localization matrix with scaling factors,
                 shape= (nparameters, nobservations)
            truncation: Define threshold for how many singular values
                 to include from SVD

        Results:
            Posterior (updated) matrix with parameters,
            shape=(nparameters, nrealizations)


        """

        N_n, N_e = Y.shape

        # Subtract the mean of every parameter, see Eqn (B.4)
        M_delta = X - np.mean(X, axis=1, keepdims=True)

        # Subtract the mean of every observation, see Eqn (B.5)
        D_delta = Y - np.mean(Y, axis=1, keepdims=True)

        # See Eqn (B.8)
        # Compute the diagonal of the inverse of S directly, without forming S itself.
        if self.C_D.ndim == 1:
            # If C_D is 1D, it's a vector of variances. S_inv_diag is 1/sqrt(variances).
            S_inv_diag = 1.0 / np.sqrt(self.C_D)
        else:
            # If C_D is 2D, extract its diagonal of variances, then compute S_inv_diag.
            S_inv_diag = 1.0 / np.sqrt(np.diag(self.C_D))

        # See Eqn (B.10)
        U, w, VT = sp.linalg.svd(S_inv_diag[:, np.newaxis] * D_delta)
        idx = singular_values_to_keep(w, truncation=truncation)
        N_r = min(N_n, N_e - 1, idx)  # Number of values in SVD to keep
        U_r, w_r = U[:, :N_r], w[:N_r]

        # See Eqn (B.12)
        # Calculate C_hat_D, the correlation matrix of measurement errors.
        # This is defined as C_hat_D = S^-1 * C_D * S^-1
        if self.C_D.ndim == 1:
            # If C_D is a 1D vector of variances,
            # it represents a diagonal matrix.
            # In this special case,
            # S_inv * C_D * S_inv simplifies to the identity matrix.
            C_hat_D = np.identity(N_n)
        else:  # C_D is a 2D matrix
            # This scales each ROW i of self.C_D by the scalar S_inv_diag[i].
            # This is numerically identical to the matrix multiplication S⁻¹ @ C_D
            C_hat_D_temp = S_inv_diag[:, np.newaxis] * self.C_D

            # This scales each COLUMN j of C_hat_D_temp by the scalar S_inv_diag[j].
            # This is numerically identical to the matrix multiplication (result) @ S⁻¹
            C_hat_D = C_hat_D_temp * S_inv_diag

        U_r_w_inv = U_r / w_r
        # See Eqn (B.13)
        R = (
            self.alpha
            * (N_e - 1)
            * np.linalg.multi_dot([U_r_w_inv.T, C_hat_D, U_r_w_inv])
        )

        # See Eqn (B.14)
        H_r, Z_r = sp.linalg.eigh(R, driver="evr", overwrite_a=True)

        # See Eqn (B.18)
        _X = (S_inv_diag[:, np.newaxis] * U_r) * (1 / w_r) @ Z_r

        # See Eqn (B.19)
        L = np.diag(1.0 / (1.0 + H_r))

        # See Eqn (B.20)
        X1 = L @ _X.T
        # See Eqn (B.21)
        X2 = D_delta.T @ _X
        # See Eqn (B.22)
        X3 = X2 @ X1

        # See Eqn (B.23)
        K_i = M_delta @ X3

        # See Eqn (B.24)
        K_rho_i = rho * K_i

        D = self.perturb_observations(ensemble_size=N_e, alpha=self.alpha)
        # See Eqn (B.25)
        X4 = K_rho_i @ (D - Y)

        # See Eqn (B.26)
        return X + X4

    def prepare_assimilation(
        self,
        Y: npt.NDArray[np.float64],
        truncation: float = 0.99,
        D: npt.NDArray[np.float64] = None,
    ):
        """
        The part of the algorithm that does not depend on the field parameters,
        but only on observations and responses are calculated here.
        No need to re-calculate this multiple times when running a loop to update
        batches of field parameters. This function implements all steps prior
        to the steps involving the field parameters and should be used as a
        preparation step once for each field parameter before running the
        loop over batches of field parameters where the field parameters
        are updated.

        Typical workflow:
          Create object of DistanceESMDA
          Prepare for update of field parameter by running 'prepare_assimilation'
          Loop over batches of field parameter values from a field parameter:
            Run 'assimilate_batch' for each batch of field parameter values


        Args:
            Y: Response matrix with predictions of observations,
               shape=(nobservations, nrealizations)
            D: Matrix of perturbed observations.
               The shape is (nobservations, nrealizations).
               Note that if D is None, the perturbed observations are drawn
               inside this function using assumption of independence
               between observation errors.

        Results:
            Updated internal matrices such that assimilate_batch function
            can be used without having to re-calculate all steps not involving
            the field parameters, but only the observations.

        """

        N_n, N_e = Y.shape

        # Subtract the mean of every observation, see Eqn (B.5)
        D_delta = Y - np.mean(Y, axis=1, keepdims=True)

        # See Eqn (B.8)
        # Compute the diagonal of the inverse of S directly, without forming S itself.
        if self.C_D.ndim == 1:
            # If C_D is 1D, it's a vector of variances. S_inv_diag is 1/sqrt(variances).
            S_inv_diag = 1.0 / np.sqrt(self.C_D)
        else:
            # If C_D is 2D, extract its diagonal of variances, then compute S_inv_diag.
            S_inv_diag = 1.0 / np.sqrt(np.diag(self.C_D))

        # See Eqn (B.10)
        U, w, VT = sp.linalg.svd(S_inv_diag[:, np.newaxis] * D_delta)
        idx = singular_values_to_keep(w, truncation=truncation)
        N_r = min(N_n, N_e - 1, idx)  # Number of values in SVD to keep
        U_r, w_r = U[:, :N_r], w[:N_r]

        # See Eqn (B.12)
        # Calculate C_hat_D, the correlation matrix of measurement errors.
        # This is defined as C_hat_D = S^-1 * C_D * S^-1
        if self.C_D.ndim == 1:
            # If C_D is a 1D vector of variances,
            # it represents a diagonal matrix.
            # In this special case,
            # S_inv * C_D * S_inv simplifies to the identity matrix.
            C_hat_D = np.identity(N_n)
        else:  # C_D is a 2D matrix
            # This scales each ROW i of self.C_D by the scalar S_inv_diag[i].
            # This is numerically identical to the matrix multiplication S⁻¹ @ C_D
            C_hat_D_temp = S_inv_diag[:, np.newaxis] * self.C_D

            # This scales each COLUMN j of C_hat_D_temp by the scalar S_inv_diag[j].
            # This is numerically identical to the matrix multiplication (result) @ S⁻¹
            C_hat_D = C_hat_D_temp * S_inv_diag

        U_r_w_inv = U_r / w_r
        # See Eqn (B.13)
        R = (
            self.alpha
            * (N_e - 1)
            * np.linalg.multi_dot([U_r_w_inv.T, C_hat_D, U_r_w_inv])
        )

        # See Eqn (B.14)
        H_r, Z_r = sp.linalg.eigh(R, driver="evr", overwrite_a=True)

        # See Eqn (B.18)
        _X = (S_inv_diag[:, np.newaxis] * U_r) * (1 / w_r) @ Z_r

        # See Eqn (B.19)
        L = np.diag(1.0 / (1.0 + H_r))

        # See Eqn (B.20)
        X1 = L @ _X.T
        # See Eqn (B.21)
        X2 = D_delta.T @ _X

        # The matrices X3 and D with perturbed observations is saved
        # and reused in assimilation of each batch of field parameters

        # See Eqn (B.22)
        self.X3 = X2 @ X1

        # Observations with added perturbations
        if D is None:
            self.D = self.perturb_observations(ensemble_size=N_e, alpha=self.alpha)
        else:
            assert self.C_D.shape[0] == D.shape[0]
            self.D = D
        return

    @memory_usage_decorator(
        enabled=ENABLE_MEMORY_LOGGING_FOR_DISTANCE_BASED_LOCALIZATION
    )
    def assimilate_batch(
        self,
        *,
        X_batch: npt.NDArray[np.float64],
        Y: npt.NDArray[np.float64],
        rho_batch: npt.NDArray[np.float64],
        D: npt.NDArray[np.float64] = None,
        truncation: float = 0.99,
    ):
        """
        Implementation of algorithm described in Appendix B of Emerick's publication.
        Reference: Emerick, Journal of Petroleum Science and Engineering 136(3):219-239
                   DOI:10.1016/j.petrol.2016.01.029  (2016)

        Calculate Ensemble Smoother update with distance-based localization using the
        RHO matrix to define the scaling/tapering function for reduction of observations
        influence on field parameters. This function implements the update steps
        of the field parameters and requires that the initial steps in the algorithm
        is already run by using the function 'prepare_assimilation'

        Args:
            X: Parameter matrix shape=(nparameters, nrealizations)
            Y: Response matrix with predictions of observations,
               shape=(nobservations, nrealizations)
            rho_batch: Localization matrix with scaling factors,
                 shape= (nparameters, nobservations)
            D: Perturbed observations, shape=(nobservations,nrealizations)
               Optional, default is to simulate the perturbations internally
               within this class

        Results:
            Posterior (updated) matrix with parameters,
            shape=(nparameters, nrealizations)


        """
        if self.X3 is None:
            # Need to run preparation step
            assert Y.shape[0] == self.observations.shape[0]
            assert Y.shape[1] == X_batch.shape[1]
            self.prepare_assimilation(Y=Y, truncation=truncation, D=D)

        # Subtract the mean of every parameter, see Eqn (B.4)
        M_delta = X_batch - np.mean(X_batch, axis=1, keepdims=True)

        # See Eqn (B.23)
        K_i = M_delta @ self.X3

        # See Eqn (B.24)
        K_rho_i = rho_batch * K_i

        # See Eqn (B.25)
        X4 = K_rho_i @ (self.D - Y)

        # See Eqn (B.26)
        return X_batch + X4

    @memory_usage_decorator(
        enabled=ENABLE_MEMORY_LOGGING_FOR_DISTANCE_BASED_LOCALIZATION
    )
    def update_params_1D(
        self,
        X_prior: npt.NDArray[np.float64],
        Y: npt.NDArray[np.float64],
        rho_1D: npt.NDArray[np.float64],
    ) -> npt.NDArray[np.float64]:
        """
        Calculate posterior update with distance-based ESMDA for one 1D parameter
        The RHO matrix is specified as input.
        Result is posterior parameter matrix

        Args:
            X_prior: Matrix with prior realizations of all field parameters,
                    shape=(nparameters, nrealizations)
            Y: Matrix with response values for each observations for each realization,
                    shape=(nobservations, nrealizations)
            rho_1D: RHO matrix elements for a 1D grid with size nx,
                    shape=(nx,nobservations). If None is input, rho is set to 1
                    for every matrix element and there will be no localization.
        Results:
            X_post: Posterior ensemble of field parameters,
            shape=(nx, nrealizations)
        """
        if Y is None or Y.shape[0] == 0:
            # No observation, no update
            return X_prior

        nobs = Y.shape[0]
        assert self.observations.shape[0] == nobs

        # Get dimensions from rho matrix
        nx, _ = rho_1D.shape
        assert nx == X_prior.shape[0]
        assert nobs == rho_1D.shape[1]
        nreal = X_prior.shape[1]
        assert Y.shape[1] == nreal

        return self.assimilate(X=X_prior, Y=Y, rho=rho_1D)

    @memory_usage_decorator(
        enabled=ENABLE_MEMORY_LOGGING_FOR_DISTANCE_BASED_LOCALIZATION
    )
    def update_params_2D(
        self,
        X_prior: npt.NDArray[np.float64],
        Y: npt.NDArray[np.float64],
        rho_2D: npt.NDArray[np.float64],
    ) -> npt.NDArray[np.float64]:
        """
        Calculate posterior update with distance-based ESMDA for one 2D parameter
        The RHO matrix is specified as input. Result is posterior parameter matrix.
        Number of parameters for the field is nx*ny,
        dimensions of the rho_2D in addition to number of observations.
        Args:
            X_prior: Matrix with prior realizations of all field parameters,
                    shape=(nparameters, nrealizations)
            Y: Matrix with response values for each observations for each realization,
                    shape=(nobservations, nrealizations)
            rho_2D: RHO matrix elements for a 2D grid with size (nx,,ny)
                    shape=(nx,ny, nobservations). If None is input, rho is set to 1
                    for every matrix element and there will be no localization.
                    nx and ny are the dimensions of the 2D field.
        Results:
            X_post: Posterior ensemble of field parameters,
                    shape=(nparameters, nrealizations)
        """
        # No update if no observations or responses
        if Y is None or Y.shape[0] == 0:
            # No update of the field parameters
            # Check if it necessary to make a copy or can we only return X_prior?
            return X_prior.copy()

        nx, ny, nobs = rho_2D.shape
        nparam = nx * ny
        nreal = X_prior.shape[1]
        assert X_prior.shape[0] == nparam, (
            f"Mismatch between X_prior dimension {X_prior.shape[0]} and nparam {nparam}"
        )

        assert nobs == Y.shape[0], (
            "Mismatch between Y matrix dimension for number of observation and rho_2D"
        )
        assert Y.shape[1] == nreal, (
            f"Mismatch between X_prior dimension {Y.shape[1]} and nreal {nreal}"
        )
        assert self.observations.shape[0] == nobs

        rho = rho_2D.reshape(nparam, nobs)
        return self.assimilate_batch(X_batch=X_prior, Y=Y, rho_batch=rho)

    @memory_usage_decorator(
        enabled=ENABLE_MEMORY_LOGGING_FOR_DISTANCE_BASED_LOCALIZATION
    )
    def update_params_3D(
        self,
        X_prior: npt.NDArray[np.float64],
        Y: npt.NDArray[np.float64],
        rho_2D_slice: npt.NDArray[np.float64],
        nz: int,
        min_nbatch: int = 1,
        no_update: bool = False,
    ) -> npt.NDArray[np.float64]:
        """
        Calculate posterior update with distance-based ESMDA for one 3D parameter.
        The RHO for one layer of the 3D field parameter is input.
        This is copied to all other layers of RHO in each batch of grid parameter
        layers since only lateral distance is used when calculating distances.
        Result is posterior parameter matrices of field parameters for one field.
        Number of parameters is nx*ny*nz where nx, ny is dimension of the rho_2D_slice
        in addition to number of observations.

        Args:
            X_prior: Matrix with prior realizations of all field parameters,
                    shape=(nparameters, nrealizations)
            Y: Matrix with response values for each observations for each realization,
                    shape=(nobservations, nrealizations)
            rho_2D_slice: RHO matrix elements for one layer (slice) of a 3D grid
                    with size (nx, ny), shape=(nx,ny,nobservations). If this variable
                    is None, then rho is set to 1 for all elements which means
                    that no localization is used.
            nz:     Number of grid layers for the 3D field parameter.
            min_nbatch: Minimum number of batches the field parameter is split into.
                    Default is 1. Usually number of batches will be calculated based
                    on available memory and the size of the field parameters,
                    number of observations and realizations. The actual number of
                    batches will be max(min_nbatch, min_number_of_batches_required).
            no_update: Is set to True only when testing the batch loop.
                    Should be False to get any update.

        Results:
            X_post: Posterior ensemble of field parameters,
            shape=(nx*ny*nz, nrealizations)
        """
        # No update if no observations or responses
        if Y is None or Y.shape[0] == 0:
            # No update of the field parameters
            # Check if it is it necessary to return a copy here?
            return X_prior.copy()

        nx, ny, nobs = rho_2D_slice.shape

        nparam_per_layer = nx * ny
        nparam = nparam_per_layer * nz
        nreal = X_prior.shape[1]
        assert X_prior.shape[0] == nparam, (
            f"Mismatch between X_prior dimension {X_prior.shape[0]} and nparam {nparam}"
        )
        assert nobs == Y.shape[0], (
            f"Mismatch between number of observations in Rho_2D {nobs} and in Y matrix"
        )
        assert Y.shape[1] == nreal, (
            f"Mismatch between X_prior dimension {Y.shape[1]} and nreal {nreal}"
        )

        X_prior_3D = X_prior.reshape(nx, ny, nz, nreal)

        # Check memory constraints and calculate how many grid layers of
        # field parameters is possible to update on one batch
        max_nlayers_per_batch = (
            calc_max_number_of_layers_per_batch_for_distance_localization(
                nx, ny, nz, nobs, nreal, bytes_per_float=8
            )
        )  # Use float64
        nlayer_per_batch = min(max_nlayers_per_batch, nz)
        nbatch = int(nz / nlayer_per_batch)

        # Number of batches is defined by available memory and wanted number of batches
        # Usually one should have as few batches as possible but sufficient to
        # be able to update a batch with the memory available.
        # It is possible to explicit require more batches than the minimum
        # number required by the memory constraint. Main use case for this
        # is probably only for unit testing to avoid having unit tests running
        # slow due to very big size of the field, the number of observations and
        # and number of realizations.
        nbatch = max(min_nbatch, nbatch)
        nlayer_per_batch = int(nz / nbatch)
        if nbatch * nlayer_per_batch == nz:
            logger.debug(f"Number of batches used: {nbatch}")
        else:
            logger.debug(f"Number of batches used: {nbatch + 1}")

        nparam_in_batch = (
            nparam_per_layer * nlayer_per_batch
        )  # For full sized batch of layers

        nlayer_last_batch = nz - nbatch * nlayer_per_batch

        # Initialize the X_post_3D
        X_post_3D = X_prior_3D.copy()
        for batch_number in range(nbatch):
            start_layer_number = batch_number * nlayer_per_batch
            end_layer_number = start_layer_number + nlayer_per_batch

            X_batch = X_prior_3D[:, :, start_layer_number:end_layer_number, :].reshape(
                (nparam_in_batch, nreal)
            )

            # Copy rho calculated from one layer of 3D parameter into all layers for
            # current batch of layers.
            # Size of rho batch: (nx,ny,nlayer_per_batch,nobs)
            rho_3D_batch = np.zeros((nx, ny, nlayer_per_batch, nobs), dtype=np.float64)
            rho_3D_batch[:, :, :, :] = rho_2D_slice[:, :, np.newaxis, :]
            rho_batch = rho_3D_batch.reshape((nparam_in_batch, nobs))

            if no_update:
                # Only used for test purpose
                X_post_batch = X_batch
            else:
                logger.debug(f"Assimilate batch number {batch_number}")
                X_post_batch = self.assimilate_batch(
                    X_batch=X_batch, Y=Y, rho_batch=rho_batch
                )
            X_post_3D[:, :, start_layer_number:end_layer_number, :] = (
                X_post_batch.reshape(nx, ny, nlayer_per_batch, nreal)
            )

        if nlayer_last_batch > 0:
            batch_number = nbatch
            start_layer_number = batch_number * nlayer_per_batch
            end_layer_number = start_layer_number + nlayer_last_batch
            nparam_in_last_batch = nparam_per_layer * nlayer_last_batch

            X_batch = X_prior_3D[:, :, start_layer_number:end_layer_number, :].reshape(
                (nparam_in_last_batch, nreal)
            )

            rho_3D_batch = np.zeros((nx, ny, nlayer_last_batch, nobs), dtype=np.float64)
            # Copy rho calculated from one layer of 3D parameter into all layers for
            # current batch of layers
            rho_3D_batch[:, :, :, :] = rho_2D_slice[:, :, np.newaxis, :]
            rho_batch = rho_3D_batch.reshape((nparam_in_last_batch, nobs))
            if no_update:
                # Only used for test purpose
                X_post_batch = X_batch
            else:
                logger.debug(f"Assimilate batch number {batch_number}")

                X_post_batch = self.assimilate_batch(
                    X_batch=X_batch, Y=Y, rho_batch=rho_batch
                )
            X_post_3D[:, :, start_layer_number:end_layer_number, :] = (
                X_post_batch.reshape(nx, ny, nlayer_last_batch, nreal)
            )

        return X_post_3D.reshape(nparam, nreal)

    def update_params(
        self,
        X_prior: npt.NDArray[np.float64],
        Y: npt.NDArray[np.float64],
        rho_input: npt.NDArray[np.float64],
        nz: int = 1,
        min_nbatch: int = 1,
        no_update: bool = False,
    ) -> npt.NDArray[np.float64]:
        """
        Calculate posterior update with distance-based ESMDA.
        Depending on the shape of rho_input it will update an 1D , 2D or 3D field.
        Dimension of the ensemble X_prior is either (nx), (nx,ny) or (nx,ny,nz)
        and rho_input define nx and ny while nz is number of layers in the grid
        for the field parameter in 3D.

        Args:
            X_prior: Matrix with prior realizations of all field parameters,
                    shape=(nparameters, nrealizations)
            Y: Matrix with response values for each observations for each realization,
                    shape=(nobservations, nrealizations)
            rho_input: RHO matrix elements.
                    If RHO_input has shape= (nx,nobs),
                    X_prior is an ensemble of 1D fields
                    If RHO_input has shape= (nx, ny, nobs),
                    X_prior is an ensemble of 2D fields.
                    If Rho_input has shape=(nx,ny,nobs) and nz > 1,
                    X_prior is an ensemble of 3D fields with size (nx, ny, nz).
            nz:     Number of grid layers for the 3D field parameter.
            min_nbatch: Minimum number of batches the field parameter is split into.
                    Default is 1. Usually number of batches will be calculated based
                    on available memory and the size of the field parameters,
                    number of observations and realizations. The actual number of
                    batches will be max(min_nbatch, min_number_of_batches_required).
            no_update: Is set to True only when testing the batch loop.
                    Should be False to get any update.

        Results:
            X_post: Posterior ensemble of field parameters,
            shape=(nx*ny*nz, nrealizations)
        """
        if rho_input.ndim == 1:
            return self.update_params_1D(X_prior=X_prior, Y=Y, rho_1D=rho_input)
        if rho_input.ndim == 2 and nz == 1:
            return self.update_params_2D(X_prior=X_prior, Y=Y, rho_2D=rho_input)
        return self.update_params_3D(
            X_prior=X_prior,
            Y=Y,
            rho_2D_slice=rho_input,
            nz=nz,
            min_nbatch=min_nbatch,
            no_update=no_update,
        )


if __name__ == "__main__":
    import pytest

    pytest.main(
        args=[
            __file__,
            "--doctest-modules",
            "-v",
        ]
    )
